#include "BrowserClient.h"
#include "BrowserApp.h"

#include <sstream>
#include <string>

#include "include/base/cef_bind.h"
#include "include/cef_app.h"
#include "include/wrapper/cef_closure_task.h"
#include "include/wrapper/cef_helpers.h"

#include "Handler/DialogHandler.h"
#include "Handler/DisplayHandler.h"
#include "Handler/LifeSpanHandler.h"
#include "Handler/LoadHandler.h"
#include "Handler/ContextMenuHandler.h"
#include "Handler/KeyboardHandler.h"
#include "Handler/RequestHandler.h"
#include "Handler/DownloadHandler.h"
#include "Handler/RenderHandler.h"
#include "Handler/JSDialogHandler.h"
#include "Handler/DragHandler.h"
#include "Handler/FocusHandler.h"


BrowserClient* BrowserClient::instance = nullptr;


BrowserClient::BrowserClient()
	: isClosing(false)
	, dialogHandler(new DialogHandler())
	, displayHandler(new DisplayHandler())
	, lifeSpanHandler(new LifeSpanHandler(*this))
	, loadHandler(new LoadHandler())
	, contextMenuHandler(new ContextMenuHandler())
	, keyboardHandler(new KeyboardHandler())
	, viewManager(new ViewManager())
	, rpcProcessor(new WebAppRPCProcessor())
	, requestHandler(new RequestHandler())
	, downloadHandler(new DownloadHandler())
	, renderHandler(new RenderHandler())
	, jsDialogHandler(new JSDialogHandler())
	, dragHandler(new DragHandler())
	, focusHandler(new FocusHandler())
	, httpRequestProvider(new HttpRequestProvider())
	, languageCode(L"en")
	, shouldQuit(false)
{
	instance = this;
}


BrowserClient::~BrowserClient()
{
	viewManager.reset();
	rpcProcessor.reset();
	httpRequestProvider.reset();
	instance = nullptr;
}


BrowserClient& BrowserClient::GetInstance()
{
	return *instance;
}


CefRefPtr<BrowserClient> BrowserClient::GetRefPtrInstance()
{
	return BrowserApp::client;
}


CefRefPtr<CefDialogHandler> BrowserClient::GetDialogHandler()
{
	return dialogHandler;
}


CefRefPtr<CefDisplayHandler> BrowserClient::GetDisplayHandler()
{
	return displayHandler;
}


CefRefPtr<CefLifeSpanHandler> BrowserClient::GetLifeSpanHandler()
{
	return lifeSpanHandler;
}


CefRefPtr<CefLoadHandler> BrowserClient::GetLoadHandler()
{
	return loadHandler;
}


CefRefPtr<CefContextMenuHandler> BrowserClient::GetContextMenuHandler()
{
	return contextMenuHandler;
}


CefRefPtr<CefKeyboardHandler> BrowserClient::GetKeyboardHandler()
{
	return keyboardHandler;
}


CefRefPtr<CefRequestHandler> BrowserClient::GetRequestHandler()
{
	return requestHandler;
}

CefRefPtr<CefDownloadHandler> BrowserClient::GetDownloadHandler()
{
	return downloadHandler;
}


CefRefPtr<CefRenderHandler> BrowserClient::GetRenderHandler()
{
	return renderHandler;
}


CefRefPtr<CefJSDialogHandler> BrowserClient::GetJSDialogHandler()
{
	return jsDialogHandler;
}


CefRefPtr<CefDragHandler> BrowserClient::GetDragHandler()
{
	return dragHandler;
}


CefRefPtr<CefFocusHandler> BrowserClient::GetFocusHandler()
{
	return focusHandler;
}


void BrowserClient::Quit()
{
	if (isClosing)
		return;

	if (!CefCurrentlyOn(TID_UI))
	{
		// Execute on the UI thread.
		CefPostTask(TID_UI, base::Bind(&BrowserClient::Quit, this));
		return;
	}

	isClosing = true;
	GetViewManager().Shutdown([&]() {
		shouldQuit.Set();
		GetRPCProcessor().Shutdown();
		CefQuitMessageLoop();
	});
}


void BrowserClient::SetLanguage(const std::wstring& languageCode)
{
	this->languageCode = languageCode;
}


std::wstring BrowserClient::GetLanguage() const
{
	return languageCode;
}


ViewManager& BrowserClient::GetViewManager()
{
	return *viewManager;
}


WebAppRPCProcessor& BrowserClient::GetRPCProcessor()
{
	return *rpcProcessor;
}


HttpRequestProvider& BrowserClient::GetHttpRequestProvider()
{
	return *httpRequestProvider;
}


ManualResetEvent& BrowserClient::GetShouldQuitEvent()
{
	return shouldQuit;
}
