#include "LifeSpanHandler.h"
#include "../BrowserClient.h"
#include "../DevToolsWebClient.h"
#include "../../Utils/CommandLine.h"
#include "../../Views/ViewManager.h"
#include "../../Views/View.h"
#include "../../Views/Components/InteractionComponent.h"
#include "../../Views/Components/EventComponent.h"
#include "../../Messages/MessageProcessor/MessageProcessor.h"
#include "../../Messages/RPCProcessor.h"
#include "../../Messages/BrowserRPCProcessor.h"

#include "include/wrapper/cef_helpers.h"
#include "include/cef_app.h"
#include "include/cef_task.h"
#include "include/base/cef_bind.h"
#include "include/wrapper/cef_closure_task.h"

#include <list>
#include <iostream>
#include <fstream> 


LifeSpanHandler::LifeSpanHandler(BrowserClient& browserClient)
	: browserClient(browserClient)
{
}


LifeSpanHandler::~LifeSpanHandler()
{
}


void LifeSpanHandler::OnAfterCreated(CefRefPtr<CefBrowser> browser)
{
	CEF_REQUIRE_UI_THREAD();

	NotifyBrowserCreated(browser);
	BrowserRPCProcessor::Add(browser, false);
}


void LifeSpanHandler::NotifyBrowserCreated(CefRefPtr<CefBrowser> browser)
{
	auto windowHandle = browser->GetHost()->GetWindowHandle();
	auto viewHandle = GetParent(windowHandle);

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(viewHandle);
	if (view != nullptr)
		view->Events.OnBrowserCreated(browser);
	else
		throw std::exception("An uncontrolled browser has been created. This should never happen.");
}


void LifeSpanHandler::ShowDevTools(CefRefPtr<CefBrowser> browser)
{
	CefPoint point;
	CefWindowInfo windowInfo;
	CefBrowserSettings settings;
	CefRefPtr<DevToolsWebClient> client = new DevToolsWebClient();

	windowInfo.SetAsPopup(browser->GetHost()->GetWindowHandle(), L"DevTools");
	browser->GetHost()->ShowDevTools(windowInfo, client, settings, point);
}


bool LifeSpanHandler::DoClose(CefRefPtr<CefBrowser> browser)
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view != nullptr)
		view->SetAllowClose();

	return false;
}


void LifeSpanHandler::OnBeforeClose(CefRefPtr<CefBrowser> browser)
{
	CEF_REQUIRE_UI_THREAD();

	BrowserRPCProcessor::Remove(browser);

	ViewManager& viewManager = browserClient.GetViewManager();
	{
		auto view = viewManager.GetView(browser);
		viewManager.RemoveView(view);
	}
}


bool LifeSpanHandler::OnBeforePopup(CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	const CefString& target_url,
	const CefString& target_frame_name,
	WindowOpenDisposition target_disposition,
	bool user_gesture,
	const CefPopupFeatures& popupFeatures,
	CefWindowInfo& windowInfo,
	CefRefPtr<CefClient>& client,
	CefBrowserSettings& settings,
	bool* no_javascript_access)
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return true;

	if (view->Events.OnCheckMustCancelRequest(frame, target_url, L"", CheckMustCancelRequestFlags::CheckSchemeAndUrl))
		return true;

	auto popupHandlingMode = view->Interaction.GetPopupHandling();
	switch (popupHandlingMode)
	{
	case PopupHandlingMode::OpenAsNewPopup:
		BrowserClient::GetInstance().GetViewManager().CreateFromPopup(view, windowInfo, browser, settings);
		return false;
	case PopupHandlingMode::OpenInCurrentView:
		browser->GetMainFrame()->LoadURL(target_url);
		return true;
	default:
		return true;
	}
}
