#include "RenderProcessHandler.h"
#include "V8/SiteKioskObjectModel.h"
#include "V8/DebugHandler.h"
#include "V8/OverrideHandler.h"
#include "../BrowserApp.h"
#include "../../Views/ViewManager.h"
#include "../../Views/View.h"
#include "../../Messages/BrowserRPCProcessor.h"

#include "include/base/cef_bind.h"
#include "include/cef_app.h"
#include "include/wrapper/cef_closure_task.h"
#include "include/wrapper/cef_helpers.h"

#include <json/json.h>
#include <fstream>

RenderProcessHandler::RenderProcessHandler()
{
}


RenderProcessHandler::~RenderProcessHandler()
{
}


void RenderProcessHandler::OnBrowserCreated(CefRefPtr<CefBrowser> browser)
{
	BrowserRPCProcessor::Add(browser, true);
	BrowserRPCProcessor::Get(browser)->GetCommands().OnRendererCreated(browser);
}


void RenderProcessHandler::OnBrowserDestroyed(CefRefPtr<CefBrowser> browser)
{
	BrowserRPCProcessor::Remove(browser);
}


void RenderProcessHandler::OnRenderThreadCreated(CefRefPtr<CefListValue> extra_info)
{
	BrowserApp::GetInstance().SetInstanceId(extra_info->GetInt(0));
}


bool RenderProcessHandler::OnBeforeNavigation(CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	CefRefPtr<CefRequest> request,
	NavigationType navigation_type,
	bool is_redirect)
{
	BrowserCommands& commands = BrowserRPCProcessor::Get(browser)->GetCommands();
	commands.OnViewIsReady();
	return commands.OnBeforeNavigation(request->GetURL(), frame->IsMain(), (int)frame->GetIdentifier());
}


void RenderProcessHandler::OnContextCreated(CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	CefRefPtr<CefV8Context> context)
{
	ApplyJavaScriptOverrideHandler(context);
	ApplyJavaScriptFullscreenFix(context);

	auto rpcProcessor = BrowserRPCProcessor::Get(browser);
	if (rpcProcessor != nullptr)
		rpcProcessor->GetCommands().OnContextCreated(frame->GetIdentifier());
}


void RenderProcessHandler::ApplyJavaScriptOverrideHandler(CefRefPtr<CefV8Context> context)
{
	auto handler = new OverrideHandler();
	CefRefPtr<CefV8Handler> handlerRefPtr = handler;
	handler->ApplyToContext(handlerRefPtr, context);
}


void RenderProcessHandler::ApplyJavaScriptFullscreenFix(CefRefPtr<CefV8Context> context)
{
	auto object = context->GetGlobal();
	auto document = object->GetValue("document");
	document->SetValue("webkitFullscreenEnabled", CefV8Value::CreateBool(false), V8_PROPERTY_ATTRIBUTE_READONLY);
	document->SetValue("webkitFullScreenEnabled", CefV8Value::CreateBool(false), V8_PROPERTY_ATTRIBUTE_READONLY);
	document->SetValue("fullscreenEnabled", CefV8Value::CreateBool(false), V8_PROPERTY_ATTRIBUTE_READONLY);
	document->SetValue("fullScreenEnabled", CefV8Value::CreateBool(false), V8_PROPERTY_ATTRIBUTE_READONLY);
}


void RenderProcessHandler::OnContextReleased(CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	CefRefPtr<CefV8Context> context)
{
	auto rpcProcessor = BrowserRPCProcessor::Get(browser);
	if (rpcProcessor != nullptr)
		rpcProcessor->GetCommands().OnContextReleased(frame->GetIdentifier());
}


bool RenderProcessHandler::OnProcessMessageReceived(CefRefPtr<CefBrowser> browser,
	CefProcessId source_process,
	CefRefPtr<CefProcessMessage> message)
{
	return false;
}


void RenderProcessHandler::RegisterDevToolsExtension()
{
#ifndef DEBUG
	std::ifstream debugTxtFile(".\\debug.txt");
	if (debugTxtFile)
#endif
	{
		std::wstring debugExtensionCode =
			L"var __openDevTools;"
			L"if (!__openDevTools)"
			L"  __openDevTools = function() {"
			L"    native function openDevTools();"
			L"    openDevTools();"
			L"  };";

		// Create an instance of my CefV8Handler object.
		CefRefPtr<CefV8Handler> debugHandler = new DebugHandler();
		CefRegisterExtension("v8/__openDevTools", debugExtensionCode, debugHandler);
	}
}

void RenderProcessHandler::RegisterSiteKioskObjectExtension()
{
	std::wstring siteKioskObjectExtensionCode =
		L"(function() {"
		L"  _siteKiosk = {};"
		L"	_siteKiosk.invokeHostRpc = function(unusedId, jsonRequestString) {"
		L"    native function invokeHostRpc();"
		L"    return invokeHostRpc(jsonRequestString);"
		L"  };"

		L"	_siteKiosk.getHasSiteKiosk = function(jsonRequestString) {"
		L"    native function getHasSiteKiosk();"
		L"    return getHasSiteKiosk();"
		L"  };"

		L"	_siteKiosk.createObjectModel = function() {"
		L"    return 0;"
		L"  };"

		L"	_siteKiosk.getSiteKioskObjectModelCode = function() {"
		L"    native function getSiteKioskObjectModelCode();"
		L"    return getSiteKioskObjectModelCode();"
		L"  };"
		L"})();";

	this->siteKioskObjectHandler = new SiteKioskObjectModel();
	CefRegisterExtension("v8/_siteKiosk", siteKioskObjectExtensionCode, siteKioskObjectHandler);
}


void RenderProcessHandler::OnWebKitInitialized()
{
	RegisterDevToolsExtension();
	RegisterSiteKioskObjectExtension();
}


void RenderProcessHandler::OnFocusedNodeChanged(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefRefPtr<CefDOMNode> node)
{
	bool isFocusOnEditableNode = node != nullptr ? node->IsEditable() : false;
	auto id = browser->GetIdentifier();
	auto pid = GetCurrentProcessId();
	BrowserRPCProcessor::Get(browser)->GetCommands().OnEditableNodeFocusChanged(isFocusOnEditableNode, frame->GetIdentifier());
}
