#include "RequestHandler.h"
#include "../BrowserClient.h"
#include "../../Views/View.h"
#include "../../Views/ViewManager.h"
#include "../../Views/Components/RequestComponent.h"
#include "../../Views/Components/EventComponent.h"
#include "../../Utils/CommandLine.h"
#include "../../Utils/StringHelper.h"


RequestHandler::RequestHandler()
	: requestFilterMode((RequestFilterMode)CommandLine::GetInstance().GetIntArgument(L"requestFilterMode", RequestFilterMode::NoRequestFilter))
{
}


RequestHandler::~RequestHandler()
{
	
}


bool RequestHandler::IsInterestingForSiteCoachSourceCheck(CefRefPtr<CefRequest> request)
{
	auto url = (std::wstring)request->GetURL();
	if (url.find(L"http") != 0) // only handle http(s)
		return false;
	return true;
}


CefRefPtr<CefResourceHandler> RequestHandler::GetResourceHandler(
	CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	CefRefPtr<CefRequest> request)
{
	if (requestFilterMode != RequestFilterMode::NoRequestFilter && IsInterestingForSiteCoachSourceCheck(request))
	{
		auto view = BrowserClient::GetInstance().GetViewManager().GetView(frame->GetBrowser());
		if (view == nullptr)
			return nullptr;

		if (view->Requests.GetContentFilterLevel() == 2)
			return new SiteCoachResourceHandler(browser, frame, request, requestFilterMode);
	}

	return nullptr;
}


bool RequestHandler::OnBeforeBrowse(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefRefPtr<CefRequest> request, bool is_redirect)
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(frame->GetBrowser());
	if (view == nullptr)
		return false;

	auto url = (std::wstring)request->GetURL();

	return view->Events.OnBeforeBrowse(url, frame->IsMain(), is_redirect, (int)frame->GetIdentifier());
}


bool RequestHandler::GetAuthCredentials(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, bool isProxy, const CefString& host, int port, const CefString& realm, const CefString& scheme, CefRefPtr<CefAuthCallback> callback)
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	callback->AddRef();
	view->Requests.RequestAuthCredentials(isProxy, host, port, realm, scheme, [callback](const AuthenticationResult& authCredentials) -> void
	{
		if (authCredentials.ShouldCancel)
		{
			callback->Cancel();
		}
		else
		{
			callback->Continue(authCredentials.Username, authCredentials.Password);
		}
		callback->Release();
	});

	return true;
}


bool RequestHandler::OnQuotaRequest(CefRefPtr<CefBrowser> browser, const CefString& origin_url, int64 new_size, CefRefPtr<CefRequestCallback> callback)
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	callback->AddRef();
	view->Requests.RequestQuota(origin_url, new_size, [callback](bool allow) -> void
	{
		callback->Continue(allow);
		callback->Release();
	});
	return true;
}


bool RequestHandler::OnCertificateError(
	CefRefPtr<CefBrowser> browser,
	cef_errorcode_t cert_error,
	const CefString& request_url,
	CefRefPtr<CefSSLInfo> ssl_info,
	CefRefPtr<CefRequestCallback> callback)
{
	callback->AddRef();
	BrowserClient::GetInstance().GetRPCProcessor().GetAppCommands().OnCertificateError(cert_error, request_url, [callback](bool allow) -> void
	{
		callback->Continue(allow);
		callback->Release();
	});
	return true;
}



CefRequestHandler::ReturnValue RequestHandler::OnBeforeResourceLoad(
	CefRefPtr<CefBrowser> browser,
	CefRefPtr<CefFrame> frame,
	CefRefPtr<CefRequest> request,
	CefRefPtr<CefRequestCallback> callback)
{
	if (request->GetMethod() == "POST")
	{
		auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
		if (view != nullptr)
		{
			auto url = request->GetURL();
			auto postData = request->GetPostData();
			if (postData != nullptr)
			{
				CefPostData::ElementVector elements;
				postData->GetElements(elements);
				std::vector<std::wstring> files;
				for (const auto& element : elements)
				{
					if (element->GetType() != CefPostDataElement::Type::PDE_TYPE_FILE)
						continue;

					auto fileName = element->GetFile();
					if (!fileName.empty())
						files.push_back(fileName);
				}
				if (!files.empty())
					if (!view->Events.OnFileUploadRequest(frame, url, files))
						return CefRequestHandler::ReturnValue::RV_CANCEL;
			}
		}
	}

	auto addUserAgent = CommandLine::GetInstance().GetStringArgument(L"addUserAgent");
	if (addUserAgent.empty())
		return RV_CONTINUE;
	addUserAgent = L"; " + addUserAgent;

	CefRequest::HeaderMap headerMap;
	request->GetHeaderMap(headerMap);

	headerMap.erase("accept-language");
	auto mainLanguage = BrowserClient::GetInstance().GetLanguage();
	auto languageParts = StringHelper::Split(mainLanguage, L'-');
	std::wstring acceptLanguage;
	if (languageParts.size() > 1)
		acceptLanguage = StringHelper::Format(L"{0}, {1};q=0.8, en;q=0.7", mainLanguage, languageParts[0]);
	else
		acceptLanguage = StringHelper::Format(L"{0}, en;q=0.7", mainLanguage);
	headerMap.insert(std::pair<CefString, CefString>("accept-language", acceptLanguage));

	auto userAgentKey = CefString("User-Agent");
	auto it = headerMap.lower_bound(userAgentKey);
	while (it != headerMap.end() && it->first == userAgentKey)
	{
		std::wstring userAgent = it->second;
		auto firstClosing = userAgent.find(L')');
		if (firstClosing == std::wstring::npos)
			continue;

		userAgent.insert(firstClosing, addUserAgent);

		it->second = userAgent;

		it++;
	}

	request->SetHeaderMap(headerMap);

	return RV_CONTINUE;
}
