#include "SiteCoachResourceHandler.h"

#include "../BrowserClient.h"
#include "../../Views/ViewManager.h"
#include "../../Views/View.h"
#include "../../Views/Components/ErrorComponent.h"
#include "../../Views/Components/EventComponent.h"
#include "../../Views/Components/RequestComponent.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/CommandLine.h"
#include "../../Utils/Uri.h"
#include "../../Http/HttpRequest.h"
#include <sstream>

#include <curl/curl.h>


SiteCoachResourceHandler::SiteCoachResourceHandler(CefRefPtr<CefBrowser> browser, CefRefPtr<CefFrame> frame, CefRefPtr<CefRequest> request, RequestFilterMode requestFilterMode)
	: frame(frame)
	, browser(browser)
	, request(request)
	, pos(0)
	, requestFilterMode(requestFilterMode)
{
}


SiteCoachResourceHandler::~SiteCoachResourceHandler()
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	if (headerCallback)
		headerCallback->Release();
	if (responseCallback)
		responseCallback->Release();
}


bool SiteCoachResourceHandler::ProcessRequest(CefRefPtr<CefRequest> request, CefRefPtr<CefCallback> callback)
{
	if (MustCancelRequestByUrl())
		return false;
	if (MustCancelRequestByFilesUploaded())
		return false;

	callback->AddRef();
	this->headerCallback = callback;

	httpRequest = BrowserClient::GetInstance().GetHttpRequestProvider().Create(request);

	httpRequest->OnRequestAuthCredentials([&](const HttpRequest& request, AuthCredentialsRequest& authCredentialsRequest)
	{
		auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
		if (view == nullptr)
		{
			authCredentialsRequest.Cancel();
			return;
		}

		view->Requests.RequestAuthCredentials(authCredentialsRequest.IsProxy, authCredentialsRequest.Host, authCredentialsRequest.Port, authCredentialsRequest.Realm, authCredentialsRequest.Scheme, [&](const AuthenticationResult& authCredentials) -> void
		{
			if (authCredentials.ShouldCancel)
				authCredentialsRequest.Cancel();
			else
				authCredentialsRequest.Continue(authCredentials.Username, authCredentials.Password);
		});
	});


	httpRequest->OnCompleted([this](const HttpRequest& request, const HttpResponse& response)
	{
		this->FinishCallback(this->headerCallback, MustCancelRequest());
	});

	httpRequest->OnAborted([this](const HttpRequest& request)
	{
		this->FinishCallback(this->headerCallback, true);
		this->FinishCallback(this->responseCallback, true);
	});

	httpRequest->OnError([this](const HttpRequest& request, cef_errorcode_t errorCode)
	{
		auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
		auto resourceType = this->request->GetResourceType();
		if (errorCode != ERR_ABORTED && view != nullptr && (resourceType == RT_SUB_FRAME || resourceType == RT_MAIN_FRAME))
			view->Events.OnLoadError(frame, errorCode, L"", this->request->GetURL());

		this->FinishCallback(this->headerCallback, true);
		this->FinishCallback(this->responseCallback, true);
	});

	httpRequest->OnHeaderDataReady([this](const HttpRequest& request, const HttpResponse& response)
	{
		// When we have no need to scan the response data we can async load the response data
		if (!MustScanRequest())
			this->FinishCallback(this->headerCallback, false);
	});

	httpRequest->Start();

	return true;
}


void SiteCoachResourceHandler::GetResponseHeaders(CefRefPtr<CefResponse> response, int64& responseLength, CefString& outRedirectUrl)
{
	const auto& responseData = httpRequest->GetResponse();
	response->SetStatus(responseData.StatusCode);

	CefResponse::HeaderMap headerMap;
	for (const auto& header : responseData.Headers)
	{
		if (responseData.StatusCode >= 300 && responseData.StatusCode <= 308 && header.first == "Location")
		{
			auto location = StringHelper::ToWString(header.second);
			if (StringHelper::StartsWith(location, std::wstring(L"/")))
			{
				auto uri = Uri::Parse(request->GetURL());
				
				auto url = (std::wstring)request->GetURL();
				url = url.substr(0, url.find(uri.Host) + uri.Host.size());
				location = url + location;
			}
			outRedirectUrl = location;
		}
		headerMap.insert(header);
	}
	response->SetHeaderMap(headerMap);

	if (!responseData.ContentType.empty())
	{
		auto elements = StringHelper::Split(responseData.ContentType, ';', 2);
		if (elements.size() >= 1)
		{
			auto mimeType = StringHelper::Trim(elements[0]);
			if (!mimeType.empty())
				response->SetMimeType(mimeType);
		}
	}

	responseLength = -1;
}


// Read response data. If data is available immediately copy up to
// |bytes_to_read| bytes into |data_out|, set |bytes_read| to the number of
// bytes copied, and return true. To read the data at a later time set
// |bytes_read| to 0, return true and call CefCallback::Continue() when the
// data is available. To indicate response completion return false.
bool SiteCoachResourceHandler::ReadResponse(void* dataOut, int bytesToRead, int& bytesRead, CefRefPtr<CefCallback> callback)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	if (bytesToRead == 0)
	{
		bytesRead = 0;
		return false;
	}

	const auto& responseData = httpRequest->GetResponse();

	if (MustScanRequest())
	{
		const auto& responseData = httpRequest->GetResponse();

		if ( (size_t)pos >= responseData.Data.Size())
		{
			bytesRead = 0;
			return false;
		}
		else
		{
			auto bytesAvailable = responseData.Data.Size() - pos;
			auto bytesReadyToWrite = min((size_t)bytesToRead, (size_t)bytesAvailable);
			std::memcpy(dataOut, &(responseData.Data.Get()[pos]), bytesReadyToWrite);
			pos += bytesReadyToWrite;
			bytesRead = bytesReadyToWrite;

			return true;
		}
	}
	else
	{
		responseCallback = callback;
		responseCallback->AddRef();

		if (httpRequest->ReadResponse(dataOut, bytesToRead, bytesRead, pos, [this](const HttpRequest& httpRequest)
		{
			this->FinishCallback(responseCallback, false);
		}))
		{
			if (responseCallback != nullptr)
				responseCallback->Release();
			responseCallback = nullptr;
			if (bytesRead == 0)
				return false;
		}
		return true;
	}
}


void SiteCoachResourceHandler::Cancel()
{
	if (httpRequest != nullptr)
		httpRequest->Cancel();
}


bool SiteCoachResourceHandler::CanGetCookie(const CefCookie& cookie)
{
	return true;
}

bool SiteCoachResourceHandler::MustScanRequest()
{
	auto type = request->GetResourceType();
	if (type == RT_IMAGE ||
		type == RT_FONT_RESOURCE ||
		type == RT_OBJECT ||
		type == RT_MEDIA ||
		type == RT_WORKER ||
		type == RT_SHARED_WORKER ||
		type == RT_FAVICON ||
		type == RT_PING ||
		type == RT_SERVICE_WORKER)
		return false;

	const auto& responseData = httpRequest->GetResponse();
	if (responseData.ContentType.empty())
		return false;

	return  (requestFilterMode == RequestFilterMode::FilterAllTextRequests && (responseData.ContentType.find("text") != responseData.ContentType.npos || responseData.ContentType.find("html") != responseData.ContentType.npos || responseData.ContentType.find("javascript") != responseData.ContentType.npos || responseData.ContentType.find("json") != responseData.ContentType.npos)) ||
		(requestFilterMode == RequestFilterMode::FilterHTMLContent && responseData.ContentType.find("html") != responseData.ContentType.npos);
}


bool SiteCoachResourceHandler::MustCancelRequest()
{
	if (!MustScanRequest())
		return false;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(frame->GetBrowser());
	if (view == nullptr)
		return false;

	auto url = (std::wstring)request->GetURL();

	const auto& responseData = httpRequest->GetResponse();
	std::string stringAContent(responseData.Data.Get(), responseData.Data.Size());
	std::wstring stringContent;
	if (responseData.ContentIsUtf8)
		stringContent = StringHelper::ToWStringNoException(stringAContent);
	else
		stringContent = std::wstring(stringAContent.begin(), stringAContent.end());

	return view->Events.OnCheckMustCancelRequest(frame, url, stringContent, CheckMustCancelRequestFlags::CheckAll);
}


bool SiteCoachResourceHandler::MustCancelRequestByUrl()
{
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(frame->GetBrowser());
	if (view == nullptr)
		return false;

	auto url = (std::wstring)request->GetURL();

	return view->Events.OnCheckMustCancelRequest(frame, url, L"", CheckMustCancelRequestFlags::CheckSchemeAndUrl);
}


bool SiteCoachResourceHandler::MustCancelRequestByFilesUploaded()
{
	if (request->GetMethod() == "POST")
	{
		auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
		if (view != nullptr)
		{
			auto url = request->GetURL();
			auto postData = request->GetPostData();
			CefPostData::ElementVector elements;
			postData->GetElements(elements);
			std::vector<std::wstring> files;
			for (const auto& element : elements)
			{
				if (element->GetType() != CefPostDataElement::Type::PDE_TYPE_FILE)
					continue;

				auto fileName = element->GetFile();
				if (!fileName.empty())
					files.push_back(fileName);
			}
			if (!files.empty())
				if (!view->Events.OnFileUploadRequest(frame, url, files))
					return true;
		}
	}
	return false;
}


void SiteCoachResourceHandler::FinishCallback(CefRefPtr<CefCallback>& callback, bool cancel)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);
	if (callback == nullptr)
		return;

	if (cancel)
		callback->Cancel();
	else
		callback->Continue();

	callback->Release();
	callback = nullptr;
}
