#include "ExceptionDump.h"

#include <dbghelp.h>


LONG WINAPI CrashHandler(EXCEPTION_POINTERS* ep)
{
	ExceptionDump dump(ep);
	dump.Dump();

	return EXCEPTION_CONTINUE_SEARCH;
}


void ExceptionDump::InitializeExceptionHandling()
{
	::SetUnhandledExceptionFilter(CrashHandler);
}

ExceptionDump::ExceptionDump(EXCEPTION_POINTERS* ep)
	: ep(ep)
{
}


ExceptionDump::~ExceptionDump()
{
}


std::wstring ExceptionDump::GenerateExceptionCodeString(DWORD code)
{
	switch (code) {
	case EXCEPTION_ACCESS_VIOLATION:         return L"EXCEPTION_ACCESS_VIOLATION";
	case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:    return L"EXCEPTION_ARRAY_BOUNDS_EXCEEDED";
	case EXCEPTION_BREAKPOINT:               return L"EXCEPTION_BREAKPOINT";
	case EXCEPTION_DATATYPE_MISALIGNMENT:    return L"EXCEPTION_DATATYPE_MISALIGNMENT";
	case EXCEPTION_FLT_DENORMAL_OPERAND:     return L"EXCEPTION_FLT_DENORMAL_OPERAND";
	case EXCEPTION_FLT_DIVIDE_BY_ZERO:       return L"EXCEPTION_FLT_DIVIDE_BY_ZERO";
	case EXCEPTION_FLT_INEXACT_RESULT:       return L"EXCEPTION_FLT_INEXACT_RESULT";
	case EXCEPTION_FLT_INVALID_OPERATION:    return L"EXCEPTION_FLT_INVALID_OPERATION";
	case EXCEPTION_FLT_OVERFLOW:             return L"EXCEPTION_FLT_OVERFLOW";
	case EXCEPTION_FLT_STACK_CHECK:          return L"EXCEPTION_FLT_STACK_CHECK";
	case EXCEPTION_FLT_UNDERFLOW:            return L"EXCEPTION_FLT_UNDERFLOW";
	case EXCEPTION_ILLEGAL_INSTRUCTION:      return L"EXCEPTION_ILLEGAL_INSTRUCTION";
	case EXCEPTION_IN_PAGE_ERROR:            return L"EXCEPTION_IN_PAGE_ERROR";
	case EXCEPTION_INT_DIVIDE_BY_ZERO:       return L"EXCEPTION_INT_DIVIDE_BY_ZERO";
	case EXCEPTION_INT_OVERFLOW:             return L"EXCEPTION_INT_OVERFLOW";
	case EXCEPTION_INVALID_DISPOSITION:      return L"EXCEPTION_INVALID_DISPOSITION";
	case EXCEPTION_NONCONTINUABLE_EXCEPTION: return L"EXCEPTION_NONCONTINUABLE_EXCEPTION";
	case EXCEPTION_PRIV_INSTRUCTION:         return L"EXCEPTION_PRIV_INSTRUCTION";
	case EXCEPTION_SINGLE_STEP:              return L"EXCEPTION_SINGLE_STEP";
	case EXCEPTION_STACK_OVERFLOW:           return L"EXCEPTION_STACK_OVERFLOW";
	default: return L"UNKNOWN EXCEPTION";
	}
}


std::wstring ExceptionDump::GetOpDescription(const ULONG opcode)
{
	switch (opcode) {
	case 0: return L"read";
	case 1: return L"write";
	case 8: return L"user-mode data execution prevention (DEP) violation";
	default: return L"unknown";
	}
}


void ExceptionDump::InitalizeSymbols() {
	// Initializes the symbols for the process.
	// Defer symbol load until they're needed, use undecorated names, and
	// get line numbers.
	SymSetOptions(SYMOPT_DEFERRED_LOADS |
		SYMOPT_UNDNAME |
		SYMOPT_LOAD_LINES);
	if (!SymInitialize(GetCurrentProcess(), NULL, TRUE)) {
		return;
	}

}

void ExceptionDump::OutputTraceToStream(const void* const* trace, size_t count, std::wstringstream& ss)
{
	for (size_t i = 0; (i < count) && ss.good(); ++i)
	{
		const int kMaxNameLength = 256;
		DWORD_PTR frame = reinterpret_cast<DWORD_PTR>(trace[i]);

		// Code adapted from MSDN example:
		// http://msdn.microsoft.com/en-us/library/ms680578(VS.85).aspx
		ULONG64 buffer[
			(sizeof(SYMBOL_INFO) +
				kMaxNameLength * sizeof(wchar_t) +
				sizeof(ULONG64) - 1) /
				sizeof(ULONG64)];
		memset(buffer, 0, sizeof(buffer));

		// Initialize symbol information retrieval structures.
		DWORD64 sym_displacement = 0;
		PSYMBOL_INFO symbol = reinterpret_cast<PSYMBOL_INFO>(&buffer[0]);
		symbol->SizeOfStruct = sizeof(SYMBOL_INFO);
		symbol->MaxNameLen = kMaxNameLength - 1;
		BOOL has_symbol = SymFromAddr(GetCurrentProcess(), frame,
			&sym_displacement, symbol);

		// Attempt to retrieve line number information.
		DWORD line_displacement = 0;
		IMAGEHLP_LINE64 line = {};
		line.SizeOfStruct = sizeof(IMAGEHLP_LINE64);
		BOOL has_line = SymGetLineFromAddr64(GetCurrentProcess(), frame,
			&line_displacement, &line);

		// Output the backtrace line.
		ss << L"    ";
		if (has_symbol)
			ss << symbol->Name << L" [0x" << trace[i] << L"+" << sym_displacement << L"]";
		else
			// If there is no symbol information, add a spacer.
			ss << L"(No symbol) [0x" << trace[i] << L"]";
		if (has_line)
			ss << L" (" << line.FileName << L":" << line.LineNumber << L")";
		ss << std::endl;
	}
}


void ExceptionDump::Dump()
{
	InitalizeSymbols();
	auto code = ep->ExceptionRecord->ExceptionCode;

	std::wstringstream ss;

	if (code == EXCEPTION_ACCESS_VIOLATION || code == EXCEPTION_IN_PAGE_ERROR)
		ss << L"Invalid operation: " << GetOpDescription(ep->ExceptionRecord->ExceptionInformation[0]) << " at address 0x" << std::hex << ep->ExceptionRecord->ExceptionInformation[1] << std::dec << std::endl;
	else
		ss << "Unhandled exception " << GenerateExceptionCodeString(code) << " at address 0x" << std::hex << ep->ExceptionRecord->ExceptionInformation[1] << std::dec << std::endl;

	// When walking an exception stack, we need to use StackWalk64().
	const int kMaxTraces = 100;
	void* trace_[kMaxTraces];
	auto count_ = 0;
	// StackWalk64() may modify context record passed to it, so we will
	// use a copy.
	CONTEXT context_record = *ep->ContextRecord;
	// Initialize stack walking.
	STACKFRAME64 stack_frame;
	memset(&stack_frame, 0, sizeof(stack_frame));

	int machine_type = IMAGE_FILE_MACHINE_I386;
	stack_frame.AddrPC.Offset = context_record.Eip;
	stack_frame.AddrFrame.Offset = context_record.Ebp;
	stack_frame.AddrStack.Offset = context_record.Esp;

	stack_frame.AddrPC.Mode = AddrModeFlat;
	stack_frame.AddrFrame.Mode = AddrModeFlat;
	stack_frame.AddrStack.Mode = AddrModeFlat;
	while (StackWalk64(machine_type,
		GetCurrentProcess(),
		GetCurrentThread(),
		&stack_frame,
		&context_record,
		NULL,
		&SymFunctionTableAccess64,
		&SymGetModuleBase64,
		NULL) &&
		count_ < kMaxTraces) {
		trace_[count_++] = reinterpret_cast<void*>(stack_frame.AddrPC.Offset);
	}

	for (size_t i = count_; i < kMaxTraces; ++i)
		trace_[i] = NULL;

	OutputTraceToStream(trace_, count_, ss);

	OutputDebugString(ss.str().c_str());
}
