#include "HttpRequest.h"
#include "HttpRequestProvider.h"
#include "../Browser/BrowserClient.h"
#include "../Utils/StringHelper.h"
#include "../Utils/CommandLine.h"
#include "../Utils/Uri.h"
#include "../Utils/UriHelper.h"
#include <string>
#include <fstream>


HttpRequest::HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, int minTLSSubVersion)
	: HttpRequest(owner, request, false, minTLSSubVersion)
{
}


HttpRequest::HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, bool ignoreSSLErrors, int minTLSSubVersion)
	: HttpRequest(owner, request, ignoreSSLErrors, minTLSSubVersion, L"", L"")
{
}


HttpRequest::HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, bool ignoreSSLErrors, int minTLSSubVersion, const std::wstring& user, const std::wstring& password)
	: baseRequest(request)
	, ignoreSSLErrors(ignoreSSLErrors)
	, curl(nullptr)
	, user(user)
	, password(password)
	, chunk(nullptr)
	, shouldCancel(false)
	, owner(owner)
	, isFinished(false)
	, tlsVersion(3)
	, minTLSSubVersion(minTLSSubVersion)
	, finishedEvent(false)
{
	requestUrl = request->GetURL();
	GenerateUserAgent();
}


HttpRequest::~HttpRequest()
{
	Cancel();
	finishedEvent.Wait();
	
	if (chunk != nullptr)
		curl_slist_free_all(chunk);
	chunk = nullptr;
	
	if (curl != nullptr)
	{
		auto uri = Uri::Parse((std::wstring)baseRequest->GetURL());
		owner.ReleaseCURLHandle(uri.Host, curl);
	}
	curl = nullptr;
}


void HttpRequest::GenerateCurlRequest()
{
	if (curl != nullptr)
	{
		response = HttpResponse();
		curl_easy_reset(curl);
	}
	else
	{
		auto uri = Uri::Parse((std::wstring)baseRequest->GetURL());
		curl = owner.GetCURLHandle(uri.Host);
	}

	if (curl)
	{
		auto url = (std::string)baseRequest->GetURL();

		curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
		curl_easy_setopt(curl, CURLOPT_USERAGENT, userAgent.c_str());
		curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "gzip,deflate");
		curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 0);
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, WriteFunc);
		curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, WriteHeaderFunc);
		curl_easy_setopt(curl, CURLOPT_XFERINFOFUNCTION, CheckAbortFunc);
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, this);
		curl_easy_setopt(curl, CURLOPT_HEADERDATA, this);
		curl_easy_setopt(curl, CURLOPT_XFERINFODATA, this);
		curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 0);
		curl_easy_setopt(curl, CURLOPT_CAINFO, owner.GetCABundlePath().c_str());

		auto tlsVersionEnum = CURL_SSLVERSION_DEFAULT;
		switch (tlsVersion)
		{
		case 2:
			tlsVersionEnum = CURL_SSLVERSION_TLSv1_2;
			break;
		case 1:
			tlsVersionEnum = CURL_SSLVERSION_TLSv1_1;
			break;
		case 0:
			tlsVersionEnum = CURL_SSLVERSION_TLSv1_0;
			break;
		default:
			tlsVersionEnum = CURL_SSLVERSION_DEFAULT;
		}

		curl_easy_setopt(curl, CURLOPT_SSLVERSION, tlsVersionEnum);

		if (ignoreSSLErrors)
		{
			curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);
			curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0);
		}

		FillHttpMethod();
		FillHeaderMap();
		FillAuthData();
		FillPostData();
	}
}


bool HttpRequest::Perform()
{
	GenerateCurlRequest();

	if (curl == nullptr)
		return true;
	auto res = curl_easy_perform(curl);

	{
		std::lock_guard<std::recursive_mutex> lock(mutex);
		response.IsCompleted = true;

		if (onResponseDataReceived)
			onResponseDataReceived(*this);
	}

	if (HandleRequestCancelation(res))
		return true;

	if (HandleCertificateErrors(res))
		return false;

	curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &response.StatusCode);
	if (HandleAuthentication())
		return false;

	if (res != CURLE_OK)
		HandleRequestError(res);
	else if (onCompleted)
		onCompleted(*this, response);

	return true;
}


int HttpRequest::CheckAbortFunc(HttpRequest *that, curl_off_t dltotal, curl_off_t dlnow, curl_off_t ultotal, curl_off_t ulnow)
{
	return that->shouldCancel ? 1 : 0;
}


size_t HttpRequest::WriteFunc(void *ptr, size_t size, size_t count, HttpRequest *that)
{
	auto realSize = size * count;
	
	std::function<void(const HttpRequest&)> func;

	{
		std::lock_guard<std::recursive_mutex> lock(that->mutex);
		that->response.Data.Append((char*)ptr, realSize);
		func = std::move(that->onResponseDataReceived);
	}

	if (func)
		func(*that);

	return realSize;
}


size_t HttpRequest::WriteHeaderFunc(void *ptr, size_t size, size_t count, HttpRequest *that)
{
	auto realSize = size * count;
	auto header = StringHelper::Trim(std::string((char*)ptr, realSize));
	if (header.empty())
	{
		// When we already have the 100 Continue statuscode, then there may follow up another header
		if (that->response.StatusCode == 100)
			return realSize;
		// header completed
		that->ExtractContentType();
		that->ExtractContentLength();

		if (that->onHeaderDataReady)
			that->onHeaderDataReady(*that, that->response);
		return realSize;
	}

	if (StringHelper::StartsWith(header, std::string("HTTP/")))
	{
		// received statuscode, clear response data
		that->response = HttpResponse();

		// extract statuscode and statustext
		auto statusElements = StringHelper::Split(header, ' ', 3);
		statusElements.erase(statusElements.begin());
		if (!statusElements.empty())
		{
			that->response.StatusCode = std::atoi(statusElements[0].c_str());
			if (statusElements.size() > 1)
				that->response.StatusText = statusElements[1];
		}


		return realSize;
	}

	auto elements = StringHelper::Split(header, ':', 2);
	if (elements.size() != 2)
		return realSize;

	std::string key = StringHelper::Trim(elements[0]);
	std::string value = StringHelper::Trim(elements[1]);

	if (key.empty() || value.empty())
		return realSize;

	that->response.Headers.insert(std::pair<std::string, std::string>(key, value));

	return realSize;
}


void HttpRequest::Start()
{
	owner.threadPool->Dispatch([this]()
	{
		this->StartPerform();
	});
}


bool HttpRequest::HandleRequestCancelation(CURLcode code)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);
	if (code == CURLE_ABORTED_BY_CALLBACK || shouldCancel)
	{
		if (this->onAborted)
			this->onAborted(*this);
		return true;
	}
	return false;
}


bool HttpRequest::HandleCertificateErrors(CURLcode code)
{
	if (code == CURLE_SSL_CONNECT_ERROR ||
		code == CURLE_SSL_CERTPROBLEM ||
		code == CURLE_SSL_CIPHER ||
		code == CURLE_SSL_CACERT ||
		code == CURLE_SSL_ISSUER_ERROR ||
		code == CURLE_PEER_FAILED_VERIFICATION)
	{
		if (ignoreSSLErrors)
		{
			finishedEvent.Set();
			if (this->onAborted)
				this->onAborted(*this);
			return true;
		}

		if (code == CURLE_SSL_CONNECT_ERROR  && tlsVersion > minTLSSubVersion)
		{
			tlsVersion--;
			this->StartPerform();
			return true;
		}

		if (onCertificateError)
		{
			auto url = (std::wstring)baseRequest->GetURL();
			auto errorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
			certificateErrorRequest = std::unique_ptr<CertificateErrorRequest>(new CertificateErrorRequest(errorCode, url, [&](bool allow)
			{
				if (allow)
				{
					this->ignoreSSLErrors = true;
					this->StartPerform();
				}
				else
				{
					finishedEvent.Set();
					if (this->onAborted)
						this->onAborted(*this);
				}
			}));
			onCertificateError(*this, *certificateErrorRequest);
		}
		return true;
	}
	return false;
}


bool HttpRequest::HandleAuthentication()
{
	if (response.StatusCode == 401)
	{
		if (onRequestAuthCredentials)
		{
			auto url = (std::wstring)baseRequest->GetURL();
			auto uri = Uri::Parse(url);

			auto isProxy = false;
			auto port = std::stoi(uri.Port);
			auto host = uri.Host;
			auto realm = url;
			auto scheme = uri.Protocol;

			authCredentialsRequest = std::unique_ptr<AuthCredentialsRequest>(new AuthCredentialsRequest(isProxy, host, port, realm, scheme, [&]()
			{
				finishedEvent.Set();
				if (this->onAborted)
					this->onAborted(*this);
			}, [&](const std::wstring& user, const std::wstring& password)
			{
				this->user = user;
				this->password = password;
				this->StartPerform();
			}));
			onRequestAuthCredentials(*this, *authCredentialsRequest);
		}
		return true;
	}

	if (!user.empty())
	{
		auto uri = Uri::Parse(baseRequest->GetURL());
		owner.domainToAuthMap[uri.Host] = user + L":" + password;
	}
	return false;
}


void HttpRequest::ExtractContentType()
{
	char *contentTypePtr;
	if (CURLE_OK == curl_easy_getinfo(curl, CURLINFO_CONTENT_TYPE, &contentTypePtr) && contentTypePtr != nullptr)
	{
		response.ContentType = std::string(contentTypePtr);
		response.ContentIsUtf8 = response.ContentType.find("UTF-8") != response.ContentType.npos;
		if (!response.ContentIsUtf8 && (response.ContentType.find("text/html") != response.ContentType.npos || response.ContentType.find("text/xhtml") != response.ContentType.npos))
			response.ContentIsUtf8 = response.ContentType.find("charset=\"utf-8\"") != response.ContentType.npos || response.ContentType.find("charset=utf-8") != response.ContentType.npos;
	}
}


void HttpRequest::ExtractContentLength()
{
	double contentLength;
	if (CURLE_OK == curl_easy_getinfo(curl, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &contentLength))
		response.ContentLength = (int)contentLength;
}


void HttpRequest::Cancel()
{
	{
		std::lock_guard<std::recursive_mutex> lock(mutex);
		shouldCancel = true;
	}
	finishedEvent.Wait();
}


void HttpRequest::StartPerform()
{
	if (Perform())
	{
		finishedEvent.Set();
		isFinished = true;
	}
}


void HttpRequest::GenerateUserAgent()
{
	// Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.81 Safari/537.36
	// Mozilla/5.0 (Windows) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36
	std::wstring userAgent(L"Mozilla/5.0 (Windows) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36");
	auto fullUserAgent = CommandLine::GetInstance().GetStringArgument(L"fullUserAgent");
	if (!fullUserAgent.empty())
		userAgent = fullUserAgent;
	auto addUserAgent = CommandLine::GetInstance().GetStringArgument(L"addUserAgent");
	if (!addUserAgent.empty())
		userAgent += L"; " + addUserAgent;
	this->userAgent = StringHelper::ToAString(userAgent);
}


void HttpRequest::FillHeaderMap()
{
	if (chunk == nullptr)
	{
		CefRequest::HeaderMap headerMap;
		baseRequest->GetHeaderMap(headerMap);
		for (const auto& pair : headerMap)
		{
			auto key = std::string(pair.first);
			auto value = std::string(pair.second);
			auto keyLower = key;
			std::transform(keyLower.begin(), keyLower.end(), keyLower.begin(), ::tolower);

			if (keyLower == "accept-language")
			{
				auto mainLanguage = BrowserClient::GetInstance().GetLanguage();
				auto languageParts = StringHelper::Split(mainLanguage, L'-');
				std::wstring acceptLanguage;
				if (languageParts.size() > 1)
					acceptLanguage = StringHelper::Format(L"{0}, {1};q=0.8, en;q=0.7", mainLanguage, languageParts[0]);
				else
					acceptLanguage = StringHelper::Format(L"{0}, en;q=0.7", mainLanguage);
				value = StringHelper::ToAString(acceptLanguage);
			}

			auto headerString = key + ": " + value;
			chunk = curl_slist_append(chunk, headerString.c_str());
		}
	}
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, chunk);
}


void HttpRequest::FillAuthData()
{
	if (!user.empty())
	{
		auto auth = StringHelper::ToAString(user + L":" + password);
		curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
		curl_easy_setopt(curl, CURLOPT_USERPWD, auth.c_str());
	}
	else
	{
		auto uri = Uri::Parse(baseRequest->GetURL());
		auto iter = owner.domainToAuthMap.find(uri.Host);
		if (iter != owner.domainToAuthMap.end())
		{
			auto auth = StringHelper::ToAString(iter->second);
			curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
			curl_easy_setopt(curl, CURLOPT_USERPWD, auth.c_str());
		}
	}
}


void HttpRequest::FillPostData()
{
	auto cefPostData = baseRequest->GetPostData();
	if (cefPostData == nullptr)
		return;

	if(postBuffer.Size() > 0)
	{
		//chunk = curl_slist_append(chunk, "Expect:");

		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, (curl_off_t)postBuffer.Size());
		curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, postBuffer.Get());
		return;
	}

	CefPostData::ElementVector postElements;
	cefPostData->GetElements(postElements);

	struct curl_httppost* last = nullptr;
	for (unsigned int i = 0; i < postElements.size(); ++i)
	{
		const auto& postElement = postElements[i];
		auto type = postElement->GetType();
		switch (type)
		{
		case PDE_TYPE_BYTES:
			{
				auto byteCount = postElement->GetBytesCount();
				auto bytes = new char[byteCount];

				postElement->GetBytes(byteCount, bytes);
				postBuffer.Append(bytes, byteCount);

				delete[] bytes;
			}
			break;
		case PDE_TYPE_FILE:
			{
				auto filePath = (std::string)postElement->GetFile();
				std::ifstream input(filePath, std::ios::binary);
				std::vector<char> buffer((
					std::istreambuf_iterator<char>(input)),
					(std::istreambuf_iterator<char>()));

				postBuffer.Append(reinterpret_cast<char*>(buffer.data()), buffer.size());
			}
			break;
		default:
			break;
		}
	}

	//chunk = curl_slist_append(chunk, "Expect:");
	curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, (curl_off_t)postBuffer.Size());
	curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, postBuffer.Get());
}


void HttpRequest::FillHttpMethod()
{
	std::string method = baseRequest->GetMethod();

	if (method == "PUT")
		curl_easy_setopt(curl, CURLOPT_PUT, 1);
	else if (method == "POST")
		curl_easy_setopt(curl, CURLOPT_POST, 1);
	else if (method == "HEAD")
		curl_easy_setopt(curl, CURLOPT_NOBODY, 1);
	else if (method == "GET")
		curl_easy_setopt(curl, CURLOPT_POST, 0);

	curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, method.c_str());
}


void HttpRequest::HandleRequestError(CURLcode errorCode)
{
	if (response.StatusCode == 0)
		response.StatusCode = 400;

	cef_errorcode_t cefErrorCode = cef_errorcode_t::ERR_FAILED;
	switch (errorCode)
	{
	case CURLE_UNSUPPORTED_PROTOCOL:    
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		response.StatusCode = 400;
		break;
	case CURLE_FAILED_INIT:            
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		response.StatusCode = 400;
		break;
	case CURLE_URL_MALFORMAT:          
		cefErrorCode = cef_errorcode_t::ERR_INVALID_URL;
		response.StatusCode = 400;
		break;
	case CURLE_NOT_BUILT_IN:            
		cefErrorCode = cef_errorcode_t::ERR_NOT_IMPLEMENTED;
		response.StatusCode = 400;
		break;
	case CURLE_COULDNT_RESOLVE_PROXY:   
		cefErrorCode = cef_errorcode_t::ERR_NAME_NOT_RESOLVED;
		response.StatusCode = 400;
		break;
	case CURLE_COULDNT_RESOLVE_HOST:   
		cefErrorCode = cef_errorcode_t::ERR_NAME_NOT_RESOLVED;
		break;
	case CURLE_COULDNT_CONNECT:         
		cefErrorCode = cef_errorcode_t::ERR_CONNECTION_FAILED;
		break;
	case CURLE_FTP_WEIRD_SERVER_REPLY: 
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_REMOTE_ACCESS_DENIED:
		cefErrorCode = cef_errorcode_t::ERR_ACCESS_DENIED;
		response.StatusCode = 401;
		break;
	case CURLE_HTTP2:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_PARTIAL_FILE:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_HTTP_RETURNED_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		response.StatusCode = 400;
		break;
	case CURLE_WRITE_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_UPLOAD_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_READ_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_OUT_OF_MEMORY:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_OPERATION_TIMEDOUT:
		cefErrorCode = cef_errorcode_t::ERR_TIMED_OUT;
		break;
	case CURLE_RANGE_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_HTTP_POST_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_CONNECT_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_BAD_DOWNLOAD_RESUME:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_FILE_COULDNT_READ_FILE:
		cefErrorCode = cef_errorcode_t::ERR_FILE_NOT_FOUND;
		response.StatusCode = 404;
		break;
	case CURLE_LDAP_CANNOT_BIND:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_LDAP_SEARCH_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_FUNCTION_NOT_FOUND:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_ABORTED_BY_CALLBACK:
		cefErrorCode = cef_errorcode_t::ERR_ABORTED;
		break;
	case CURLE_BAD_FUNCTION_ARGUMENT:
		cefErrorCode = cef_errorcode_t::ERR_INVALID_ARGUMENT;
		break;
	case CURLE_INTERFACE_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_TOO_MANY_REDIRECTS:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_UNKNOWN_OPTION:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_TELNET_OPTION_SYNTAX:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_PEER_FAILED_VERIFICATION:
		cefErrorCode = cef_errorcode_t::ERR_CERT_INVALID;
		break;
	case CURLE_GOT_NOTHING:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_ENGINE_NOTFOUND:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_SSL_ENGINE_SETFAILED:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_SEND_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_RECV_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_CERTPROBLEM:
		cefErrorCode = cef_errorcode_t::ERR_CERT_INVALID;
		break;
	case CURLE_SSL_CIPHER:
		cefErrorCode = cef_errorcode_t::ERR_SSL_VERSION_OR_CIPHER_MISMATCH;
		break;
	case CURLE_SSL_CACERT:
		cefErrorCode = cef_errorcode_t::ERR_SSL_CLIENT_AUTH_CERT_NEEDED;
		break;
	case CURLE_BAD_CONTENT_ENCODING:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_LDAP_INVALID_URL:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_FILESIZE_EXCEEDED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_USE_SSL_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SEND_FAIL_REWIND:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_SSL_ENGINE_INITFAILED:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_LOGIN_DENIED:
		cefErrorCode = cef_errorcode_t::ERR_ACCESS_DENIED;
		break;
	case CURLE_REMOTE_DISK_FULL:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_REMOTE_FILE_EXISTS:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_CONV_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_CONV_REQD:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_CACERT_BADFILE:
		cefErrorCode = cef_errorcode_t::ERR_CERT_INVALID;
		break;
	case CURLE_REMOTE_FILE_NOT_FOUND:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSH:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_SHUTDOWN_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_AGAIN:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_SSL_CRL_BADFILE:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_SSL_ISSUER_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_SSL_PROTOCOL_ERROR;
		break;
	case CURLE_RTSP_CSEQ_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_RTSP_SESSION_ERROR:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_CHUNK_FAILED:
		cefErrorCode = cef_errorcode_t::ERR_FAILED;
		break;
	case CURLE_NO_CONNECTION_AVAILABLE:
		cefErrorCode = cef_errorcode_t::ERR_INTERNET_DISCONNECTED;
		break;
	case CURLE_SSL_PINNEDPUBKEYNOTMATCH:
		cefErrorCode = cef_errorcode_t::ERR_CERT_INVALID;
		break;
	case CURLE_SSL_INVALIDCERTSTATUS:
		cefErrorCode = cef_errorcode_t::ERR_CERT_INVALID;
		break;
	default:
		break;
	}

	if (onError)
		onError(*this, cefErrorCode);
}


bool HttpRequest::ReadResponse(void* dataOut, int bytesToRead, int& bytesRead, int &pos, const std::function<void(const HttpRequest&)>& callback)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	bytesRead = response.Data.Consume((char*)dataOut, bytesToRead);
	
	if (bytesRead == 0)
	{
		if (!response.IsCompleted)
			onResponseDataReceived = callback;
		return response.IsCompleted;
	}
	return true;
}
