#pragma once

#include <functional>
#include <mutex>
#include "include\cef_resource_handler.h"
#include "include/cef_client.h"
#include <curl/curl.h>
#include <atomic>

#include "../Thread/ThreadPool.h"
#include "../Thread/ManualResetEvent.h"
#include "../Buffer.h"


class HttpRequestProvider;


struct HttpResponse
{
	HttpResponse()
		: StatusCode(0)
		, ContentIsUtf8(false)
		, ContentLength(-1)
		, IsCompleted(false)
	{
	}

	~HttpResponse()
	{
	}

	std::multimap<std::string, std::string> Headers;
	Buffer<char> Data;
	int ContentLength;
	int StatusCode;
	bool IsCompleted;
	std::string StatusText;
	std::string ContentType;
	bool ContentIsUtf8;
};

class CertificateErrorRequest
{
public:
	CertificateErrorRequest(int errorCode, const std::wstring& url, const std::function<void(bool)>& callback)
		: callback(callback)
		, ErrorCode(errorCode)
		, Url(url)
	{
	}

	void Continue()
	{
		callback(true);
	}

	void Cancel()
	{
		callback(false);
	}

	int ErrorCode;
	std::wstring Url;
private:
	
	std::function<void(bool)> callback;
};

class AuthCredentialsRequest
{
public:
	AuthCredentialsRequest(bool isProxy, const std::wstring& host, int port, const std::wstring& realm, const std::wstring& scheme, const std::function<void()>& cancelCallback, const std::function<void(const std::wstring&, const std::wstring&)>& continueCallback)
		: cancelCallback(cancelCallback)
		, continueCallback(continueCallback)
		, IsProxy(isProxy)
		, Host(host)
		, Port(port)
		, Realm(realm)
		, Scheme(scheme)
	{
	}

	void Continue(const std::wstring& user, const std::wstring& password)
	{
		continueCallback(user, password);
	}

	void Cancel()
	{
		cancelCallback();
	}

public:
	bool IsProxy;
	std::wstring Host;
	int Port;
	std::wstring Realm;
	std::wstring Scheme;

private:
	std::function<void()> cancelCallback;
	std::function<void(const std::wstring&, const std::wstring&)> continueCallback;
};


class HttpRequest
{
public:
	HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, int minTLSSubVersion);
	HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, bool ignoreSSLErrors, int minTLSSubVersion);
	HttpRequest(HttpRequestProvider& owner, CefRefPtr<CefRequest> request, bool ignoreSSLErrors, int minTLSSubVersion, const std::wstring& user, const std::wstring& password);
	~HttpRequest();

	void Start();
	void Cancel();

	void SetAuthCredentials(const std::wstring& user, const std::wstring& password);

	void OnHeaderDataReady(const std::function<void(const HttpRequest&, const HttpResponse&)>& callback)
	{
		onHeaderDataReady = callback;
	}

	void OnCompleted(const std::function<void(const HttpRequest&, const HttpResponse&)>& callback)
	{ 
		onCompleted = callback; 
	}

	void OnAborted(const std::function<void(const HttpRequest&)>& callback)
	{
		onAborted = callback;
	}

	void OnError(const std::function<void(const HttpRequest&, cef_errorcode_t)>& callback)
	{
		onError = callback;
	}

	void OnCertificateError(const std::function<void(const HttpRequest&, CertificateErrorRequest&)>& callback)
	{ 
		onCertificateError = callback;
	}

	void OnRequestAuthCredentials(const std::function<void(const HttpRequest&, AuthCredentialsRequest&)>& callback)
	{
		onRequestAuthCredentials = callback;
	}

	bool ReadResponse(void* dataOut, int bytesToRead, int& bytesRead, int& pos, const std::function<void(const HttpRequest&)>& callback);

	const HttpResponse& GetResponse() const { return response; }

private:
	void HandleRequestError(CURLcode errorCode);
	bool Perform();
	void StartPerform();
	void GenerateCurlRequest();
	bool HandleRequestCancelation(CURLcode code);
	bool HandleCertificateErrors(CURLcode code);
	bool HandleAuthentication();
	void ExtractContentType();
	void ExtractContentLength();
	void GenerateUserAgent();
	void FillHeaderMap();
	void FillAuthData();
	void FillPostData();
	void FillHttpMethod();

private:
	static size_t WriteFunc(void *ptr, size_t size, size_t nmemb, HttpRequest *that);
	static size_t WriteHeaderFunc(void *ptr, size_t size, size_t nmemb, HttpRequest *that);
	static int CheckAbortFunc(HttpRequest *that, curl_off_t dltotal, curl_off_t dlnow, curl_off_t ultotal, curl_off_t ulnow);

private:
	std::wstring user;
	std::wstring password;

	HttpResponse response;
	CefRefPtr<CefRequest> baseRequest;
	bool ignoreSSLErrors;
	int tlsVersion;
	int minTLSSubVersion;

	struct curl_slist *chunk;
	CURL *curl;

	std::function<void(const HttpRequest&, const HttpResponse&)> onCompleted;
	std::function<void(const HttpRequest&, const HttpResponse&)> onHeaderDataReady;
	std::function<void(const HttpRequest&)> onAborted;
	std::function<void(const HttpRequest&, CertificateErrorRequest&)> onCertificateError;
	std::function<void(const HttpRequest&, AuthCredentialsRequest&)> onRequestAuthCredentials;
	std::function<void(const HttpRequest&, cef_errorcode_t)> onError;
	std::function<void(const HttpRequest&)> onResponseDataReceived;

	std::unique_ptr<AuthCredentialsRequest> authCredentialsRequest;
	std::unique_ptr<CertificateErrorRequest> certificateErrorRequest;

	std::recursive_mutex mutex;
	ManualResetEvent finishedEvent;
	HttpRequestProvider& owner;

	volatile bool shouldCancel;
	std::string userAgent;
	bool isFinished;

	Buffer<char> postBuffer;
	std::string requestUrl;
};
