#include "HttpRequestProvider.h"
#include "HttpRequest.h"
#include "../Browser/BrowserClient.h"
#include "../Messages/RPCProcessor.h"
#include "../Messages/Commands/AppCommands.h"
#include "../Utils/Uri.h"
#include "../Utils/CommandLine.h"
#include "CABundleGenerator.h"


HttpRequestProvider::HttpRequestProvider()
	: threadPool(std::unique_ptr<ThreadPool>(new ThreadPool("HttpRequest", 0, 4)))
	, minTLSSubVersion(CommandLine::GetInstance().GetIntArgument(L"minTLSSubVersion", 0))
{
	char tempPathBuffer[MAX_PATH];
	GetTempPathA(MAX_PATH, tempPathBuffer);
	caBundlePath = std::string(tempPathBuffer) + "SiteKioskNGCef-CA.pem";

	CABundleGenerator generator(caBundlePath);
	generator.Generate();
}


HttpRequestProvider::~HttpRequestProvider()
{
}


bool HttpRequestProvider::CanNavigateToUnsafeUrl(const std::wstring& url)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	auto uri = Uri::Parse(url);

	if (std::find(allowedCertificateErrorDomains.begin(), allowedCertificateErrorDomains.end(), uri.Host) != allowedCertificateErrorDomains.end())
		return true;
	return false;
}


std::shared_ptr<PendingCertificateRequests> HttpRequestProvider::ProvidePendingCertificateRequests(const std::wstring& url, bool& isNew)
{
	auto uri = Uri::Parse(url);

	auto iter = pendingCertificateErrorRequests.find(uri.Host);
	std::shared_ptr<PendingCertificateRequests> pendingRequests;

	if (iter != pendingCertificateErrorRequests.end())
	{
		pendingRequests = iter->second;
		isNew = false;
	}
	else
	{
		isNew = true;
		pendingRequests = std::shared_ptr<PendingCertificateRequests>(new PendingCertificateRequests());
		pendingCertificateErrorRequests.insert(std::make_pair(uri.Host, pendingRequests));
	}

	return pendingRequests;
}


void HttpRequestProvider::MakeCertificateErrorRequest(std::shared_ptr<PendingCertificateRequests> pendingRequests, int errorCode, const std::wstring& url)
{
	pendingRequests->Cancelable = BrowserClient::GetInstance().GetRPCProcessor().GetAppCommands().OnCertificateError(errorCode, url, [&, pendingRequests, url](bool allow) -> void
	{
		auto uri = Uri::Parse(url);

		if (allow)
		{
			{
				std::lock_guard<std::recursive_mutex> lock(mutex);
				allowedCertificateErrorDomains.push_back(uri.Host);
			}
			pendingRequests->Continue();
		}
		else
			pendingRequests->Cancel();

		{
			std::lock_guard<std::recursive_mutex> lock(mutex);
			pendingCertificateErrorRequests.erase(pendingCertificateErrorRequests.find(uri.Host));
		}
	});

	//pendingRequests->Cancelable->OnCancel([&, url]()
	//{
	//	auto uri = Uri::Parse(url);
	//	std::lock_guard<std::recursive_mutex> lock(mutex);
	//	pendingCertificateErrorRequests.erase(pendingCertificateErrorRequests.find(uri.Host));
	//});
}


std::shared_ptr<HttpRequest> HttpRequestProvider::Create(CefRefPtr<CefRequest> request)
{
	auto httpRequest = std::shared_ptr<HttpRequest>(new HttpRequest(*this, request, minTLSSubVersion));

	httpRequest->OnCertificateError([this](const HttpRequest& httpRequest, CertificateErrorRequest& certificateErrorRequest)
	{
		if (CanNavigateToUnsafeUrl(certificateErrorRequest.Url))
		{
			certificateErrorRequest.Continue();
			return;
		}

		std::lock_guard<std::recursive_mutex> lock(mutex);

		bool isNew;
		auto pendingRequests = ProvidePendingCertificateRequests(certificateErrorRequest.Url, isNew);

		PendingCertificateRequests::PendingCertificateRequest newRequest{ certificateErrorRequest };
		pendingRequests->Requests.push_back(newRequest);

		if (isNew)
			MakeCertificateErrorRequest(pendingRequests, certificateErrorRequest.ErrorCode, certificateErrorRequest.Url);
	});

	return httpRequest;
}


CURL* HttpRequestProvider::GetCURLHandle(const std::wstring& host)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	CheckOrphanedCURLHandles();

	auto& data = hostToUnusedCurlHandles[host];
	if (data.empty())
		return curl_easy_init();
	else
	{
		auto store = data.back();
		data.pop_back();

		return store.Handle;
	}
}


void HttpRequestProvider::ReleaseCURLHandle(const std::wstring& host, CURL* handle)
{
	std::lock_guard<std::recursive_mutex> lock(mutex);

	curl_easy_reset(handle);
	StoredCurlHandle store;
	store.Handle = handle;
	store.StoredAtTime = GetTickCount();
	hostToUnusedCurlHandles[host].push_back(store);
}


void HttpRequestProvider::CheckOrphanedCURLHandles()
{
	auto currentTick = GetTickCount();
	for (auto iter = hostToUnusedCurlHandles.begin(); iter != hostToUnusedCurlHandles.end();)
	{
		for (unsigned int i = 0; i < iter->second.size();)
		{
			auto& store = iter->second[i];
			if (currentTick - store.StoredAtTime > 60 * 1000)
			{
				curl_easy_cleanup(store.Handle);
				iter->second.erase(iter->second.begin() + i);
			}
			else
				++i;
		}
		if (iter->second.size() == 0)
			iter = hostToUnusedCurlHandles.erase(iter);
		else
			++iter;
	}
}


std::string HttpRequestProvider::GetCABundlePath() const
{
	return caBundlePath;
}