#include "BrowserRPCProcessor.h"
#include "../Browser/BrowserApp.h"
#include "../Utils/StringHelper.h"
#include "../Utils/CommandLine.h"
#include "../Utils/Handle.h"
#include "../Utils/Profiler.h"
#include "../resource.h"

#include <json/json.h>
#include <sstream>
#include <fstream>


std::map<int, std::shared_ptr<CreationCount<BrowserRPCProcessor> > > BrowserRPCProcessor::browserToIPC;
const std::wstring BrowserRPCProcessor::CommandPrefix = L"Command:";

static std::wstring GenerateSenderName(CefRefPtr<CefBrowser> browser, bool isRenderProcess)
{
	std::wstringstream stream;
	stream << BrowserApp::GetInstance().GetInstanceId();
	stream << L":BrowserRPCProcessor:";
	stream << browser->GetIdentifier();
	stream << (isRenderProcess ? L":Browser" : L":Renderer");
	return stream.str();
}


static std::wstring GenerateReceiverName(CefRefPtr<CefBrowser> browser, bool isRenderProcess)
{
	std::wstringstream stream;
	stream << BrowserApp::GetInstance().GetInstanceId();
	stream << L":BrowserRPCProcessor:";
	stream << browser->GetIdentifier();
	stream << (isRenderProcess ? L":Renderer" : L":Browser");

	bool exists = false;
#ifndef USE_SHAREDMEMORY_IPC
	Handle pipe = CreateFile((L"\\\\.\\pipe\\" + stream.str()).c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (pipe)
		exists = true;
	else
		exists = GetLastError() == ERROR_PIPE_BUSY;
#else
	exists = IPCWriter::CanOpen(stream.str());
#endif
	if (exists)
		stream << ":" << GetCurrentProcessId();

	return stream.str();
}


BrowserRPCProcessor::BrowserRPCProcessor(CefRefPtr<CefBrowser> browser, bool isRenderProcess)
	: RPCProcessor()
#ifndef USE_SHAREDMEMORY_IPC
	, messageProcessor(new NetPipeMessageProcessor(GenerateSenderName(browser, isRenderProcess), GenerateReceiverName(browser, isRenderProcess)))
#else
	, messageProcessor(new SharedMemoryMessageProcessor(GenerateSenderName(browser, isRenderProcess), GenerateReceiverName(browser, isRenderProcess)))
#endif
	, browser(browser)
	, commands(*this, browser)
	, isRenderProcess(isRenderProcess)
{
	messageProcessor->AddMessageHandler(BrowserRPCProcessor::CommandPrefix, std::bind(&BrowserRPCProcessor::OnMessage, this, std::placeholders::_1));
	commands.Init();
	messageProcessor->Start();

	if (isRenderProcess)
		commands.ChangeCurrentRenderProcess(GetReceiverName());
}


BrowserRPCProcessor::~BrowserRPCProcessor()
{
	Shutdown();
}

bool BrowserRPCProcessor::GetIsBrowser() const
{
	return !isRenderProcess;
}


void BrowserRPCProcessor::Shutdown()
{
	std::unique_lock<std::mutex> lock(mutex);
	RPCProcessor::Shutdown();
	if (messageProcessor != nullptr)
	{
		messageProcessor->Shutdown();
		messageProcessor.reset();
	}
}


void BrowserRPCProcessor::Add(CefRefPtr<CefBrowser> browser, bool isRenderProcess)
{
	auto id = browser->GetIdentifier();
	auto iter = browserToIPC.find(id);

	if (iter == browserToIPC.end())
	{
		std::shared_ptr<BrowserRPCProcessor > rpcProcessor(new BrowserRPCProcessor(browser, isRenderProcess));
		std::shared_ptr<CreationCount<BrowserRPCProcessor> > creationCount(new CreationCount<BrowserRPCProcessor>(rpcProcessor));
		browserToIPC.insert(std::make_pair(browser->GetIdentifier(), creationCount));
	}
	else
	{
		iter->second->Data->Shutdown();
		std::shared_ptr<BrowserRPCProcessor > rpcProcessor(new BrowserRPCProcessor(browser, isRenderProcess));
		iter->second->Set(rpcProcessor);
	}
}


void BrowserRPCProcessor::Remove(CefRefPtr<CefBrowser> browser)
{
	if (browser == nullptr)
		return;

	auto id = browser->GetIdentifier();
	auto iter = browserToIPC.find(id);
	if (iter == browserToIPC.end())
		return;

	if (iter->second->TryRemove([](std::shared_ptr<BrowserRPCProcessor> d)
	{
		d->Shutdown();
	}))
		browserToIPC.erase(iter);
}


std::shared_ptr<BrowserRPCProcessor> BrowserRPCProcessor::Get(CefRefPtr<CefBrowser> browser)
{
	if (browser == nullptr)
		return nullptr;

	auto id = browser->GetIdentifier();
	auto iter = browserToIPC.find(id);
	if (iter == browserToIPC.end())
		return nullptr;

	return iter->second->Data;
}


void BrowserRPCProcessor::CreateNewSendTarget(const std::wstring name)
{
	/*if (messageProcessor->GetSendTarget() == name)
		return;*/

	messageProcessor->CreateNewSendTarget(name);
}


void BrowserRPCProcessor::RegisterCommand(const std::wstring& method, RPCFunction commandFunction)
{
	std::unique_lock<std::mutex> lock(mutex);

	Handler handler(method, commandFunction);
	registeredCommands.push_back(handler);
}


bool BrowserRPCProcessor::OnMessage(const std::wstring& message)
{
	PERFORMANCER;

	auto values = StringHelper::Split(message, L':', 2);
	int targetId = std::stoi(values[0]);
	auto realMessage = values[1];

	// Dont handle messages send for an other render process
	auto processId = GetCurrentProcessId();
	if (isRenderProcess && targetId != 0 && processId != targetId)
		return true;

	JsonHandlerInfo jsonHandlerInfo;
	jsonHandlerInfo.Handlers = registeredCommands;
	jsonHandlerInfo.ResponsePrefix = BrowserRPCProcessor::CommandPrefix + std::to_wstring(targetId) + L":";

	return HandleJsonRequest(jsonHandlerInfo, realMessage, targetId);
}
