#include "AppCommands.h"
#include "../WebAppRPCProcessor.h"
#include "../MessageProcessor/MessageProcessor.h"
#include "../../Browser/BrowserClient.h"
#include "../../Views/ViewManager.h"
#include "../../Views/View.h"
#include "../../Views/Components/DisplayComponent.h"
#include "../../Utils/StringHelper.h"

#include <json/json.h>
#include <functional>

AppCommands::AppCommands(WebAppRPCProcessor& rpcProcessor)
	: rpcProcessor(rpcProcessor)
{

}


void AppCommands::Init()
{
	rpcProcessor.RegisterAppCommand(L"createEmptyView", std::bind(&AppCommands::CreateEmptyViewHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterAppCommand(L"createView", std::bind(&AppCommands::CreateViewHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterAppCommand(L"quit", std::bind(&AppCommands::QuitHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterAppCommand(L"ping", std::bind(&AppCommands::PingHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterAppCommand(L"setLanguage", std::bind(&AppCommands::SetLanguageHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterAppCommand(L"setKeyboardLayout", std::bind(&AppCommands::SetKeyboardLayoutHandler, this, std::placeholders::_1, std::placeholders::_2));
}


bool AppCommands::CreateEmptyViewHandler(const Json::Value& params, Json::Value& response)
{
	auto view = BrowserClient::GetInstance().GetViewManager().Create();

	response["result"] = view->GetId();

	return true;
}


bool AppCommands::CreateViewHandler(const Json::Value& params, Json::Value& response)
{
	auto options = params[0];

	CreateViewParams createViewParams;

	createViewParams.Left = options.get("left", 0).asInt();
	createViewParams.Top = options.get("top", 0).asInt();
	createViewParams.Width = options.get("width", 640).asInt();
	createViewParams.Height = options.get("height", 480).asInt();
	createViewParams.ShowBorders = options.get("showBorders", true).asBool();
	createViewParams.IsClosable = options.get("isClosable", true).asBool();
	createViewParams.StartUrl = StringHelper::ToWString(options.get("url", "about:blank").asString());
	createViewParams.ParentId = options.get("parentId", -1).asInt();
	createViewParams.AllowInteraction = options.get("allowInteraction", true).asBool();
	createViewParams.IsVisible = options.get("isVisible", true).asBool();
	createViewParams.ContentFilterLevel = options.get("contentFilterLevel", true).asBool();

	auto view = BrowserClient::GetInstance().GetViewManager().Create(createViewParams);

	Json::Value result;
	result[0] = view->GetId();
	result[1] = (int)view->GetHandle();

	response["result"] = result;

	return true;
}


bool AppCommands::QuitHandler(const Json::Value& params, Json::Value& response)
{
	BrowserClient::GetInstance().Quit();
	return true;
}


bool AppCommands::PingHandler(const Json::Value& params, Json::Value& response)
{
	BrowserClient::GetInstance().GetViewManager().Ping();
	return true;
}


bool AppCommands::SetLanguageHandler(const Json::Value& params, Json::Value& response)
{
	BrowserClient::GetInstance().SetLanguage(StringHelper::ToWString(params[0].asString()));
	return true;
}


bool AppCommands::SetKeyboardLayoutHandler(const Json::Value& params, Json::Value& response)
{
	HKL keyboardLayoutHandle = (HKL)params[0].asInt();
	::ActivateKeyboardLayout(keyboardLayoutHandle, KLF_SETFORPROCESS);
	return true;
}



std::wstring AppCommands::GetString(const std::wstring& stringId)
{
	Json::Value params;
	params[0] = StringHelper::ToAString(stringId);

	auto response = rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"getString")
		.UseParameters(params)
		.RunAndWait();

	return StringHelper::ToWString(response["result"].asString());
}


bool AppCommands::MustApplyObjectModel(const std::wstring& url, HANDLE exitEventHandle)
{
	Json::Value params;
	params[0] = StringHelper::ToAString(url);

	auto response = rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"mustApplyObjectModel")
		.UseParameters(params)
		.RunAndWait(exitEventHandle);

	return response["result"].asBool();
}


void AppCommands::CancelRequest(int requestId)
{
	Json::Value params;
	params[0] = requestId;

	rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"cancelRequest")
		.UseParameters(params)
		.Run();
}


void AppCommands::OnAppReady()
{
	rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"ready")
		.Run();
}


std::shared_ptr<Cancelable> AppCommands::OnCertificateError(int certError, const std::wstring& requestUrl, const std::function<void(bool)>& callback)
{
	Json::Value params;
	params[0] = certError;
	params[1] = StringHelper::ToAString(requestUrl);
	params[2] = (int)BrowserClient::GetInstance().GetViewManager().GetView(1)->GetHandle();

	auto onResponse = [callback](const Json::Value& response) -> void
	{
		callback(response["result"].asBool());
	};

	auto cancelable = rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"certificateError")
		.UseParameters(params)
		.RunAsync(onResponse);

	auto requestId = cancelable->GetId();
	cancelable->OnCancel([&, callback, requestId]() -> void {
		callback(false);
		CancelRequest(requestId);
	}, true);

	return cancelable;
}


void AppCommands::OnEditableNodeFocusChanged(bool isEditable)
{
	Json::Value params;
	params[0] = isEditable;

	rpcProcessor.CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"editableNodeFocusChanged")
		.UseParameters(params)
		.Run();
}
