#include "BrowserCommands.h"
#include "../BrowserRPCProcessor.h"
#include "../MessageProcessor/MessageProcessor.h"
#include "../../Browser/BrowserClient.h"
#include "../../Browser/BrowserApp.h"
#include "../../Messages/RPCProcessor.h"
#include "../../Browser/DevToolsWebClient.h"
#include "../../Browser/BrowserApp.h"
#include "../../Views/ViewManager.h"
#include "../../Views/View.h"
#include "../../Views/Components/InteractionComponent.h"
#include "../../Views/Components/EventComponent.h"
#include "../../Views/Components/PingComponent.h"
#include "../../Utils/UriHelper.h"
#include "../../Utils/Profiler.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/CommandLine.h"

#include "include/base/cef_bind.h"
#include "include/cef_app.h"
#include "include/wrapper/cef_closure_task.h"
#include "include/wrapper/cef_helpers.h"

#include <json/json.h>
#include <functional>
#include <iostream>


BrowserCommands::BrowserCommands(BrowserRPCProcessor& rpcProcessor, CefRefPtr<CefBrowser> browser)
	: rpcProcessor(rpcProcessor)
	, browser(browser)
	, viewIsReady(false)
{

}


void BrowserCommands::Init()
{
	rpcProcessor.RegisterCommand(L"beforeNavigation", std::bind(&BrowserCommands::OnBeforeNavigationHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"contextCreated", std::bind(&BrowserCommands::OnContextCreatedHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"contextReleased", std::bind(&BrowserCommands::OnContextReleasedHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"viewIsReady", std::bind(&BrowserCommands::OnViewIsReadyHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"changeCurrentRenderProcess", std::bind(&BrowserCommands::OnChangeCurrentRenderProcess, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"rendererCreated", std::bind(&BrowserCommands::OnRendererCreatedHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"executeRPC", std::bind(&BrowserCommands::OnExecuteRPCHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"ping", std::bind(&BrowserCommands::OnPingHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"pong", std::bind(&BrowserCommands::OnPongHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"openDevTools", std::bind(&BrowserCommands::OnOpenDevToolsHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"mustApplyObjectModel", std::bind(&BrowserCommands::OnMustApplyObjectModelHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"editableNodeFocusChanged", std::bind(&BrowserCommands::OnEditableNodeFocusChangedHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"print", std::bind(&BrowserCommands::OnPrintHandler, this, std::placeholders::_1, std::placeholders::_2));
	rpcProcessor.RegisterCommand(L"getAssetsPath", std::bind(&BrowserCommands::OnGetAssetsPath, this, std::placeholders::_1, std::placeholders::_2));
}


std::wstring BrowserCommands::ExecuteRPC(const std::wstring& rpc)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = StringHelper::ToAString(rpc);

	auto response = rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"executeRPC")
		.UseParameters(params)
		.RunAndWait();

	Json::FastWriter writer;
	auto jsonString = writer.write(response["result"]);
	jsonString.pop_back();

	return StringHelper::ToWString(jsonString);
}


void BrowserCommands::Ping()
{
	PERFORMANCER;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"ping")
		.Run();
}


void BrowserCommands::Pong()
{
	PERFORMANCER;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"pong")
		.Run();
}


void BrowserCommands::OpenDevTools()
{
	PERFORMANCER;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"openDevTools")
		.Run();
}


void BrowserCommands::OnViewIsReady()
{
	PERFORMANCER;

	if (viewIsReady)
		return;
	viewIsReady = true;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"viewIsReady")
		.Run();
}


void BrowserCommands::ChangeCurrentRenderProcess(const std::wstring name)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = StringHelper::ToAString(name);
	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"changeCurrentRenderProcess")
		.UseParameters(params)
		.RunAndWait();
}


bool BrowserCommands::MustApplyObjectModel(const std::wstring& url)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = StringHelper::ToAString(url);

	auto response = rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"mustApplyObjectModel")
		.UseParameters(params)
		.RunAndWait();

	return response["result"].asBool();
}


void BrowserCommands::Print()
{
	PERFORMANCER;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"print")
		.RunAndWait();
}


std::wstring BrowserCommands::GetAssetsPath() const
{
	PERFORMANCER;

	auto response = rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"getAssetsPath")
		.RunAndWait();

	return StringHelper::ToWString(response["result"].asString());
}


void BrowserCommands::OnRendererCreated(CefRefPtr<CefBrowser> browser)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = browser->GetIdentifier();

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"rendererCreated")
		.UseParameters(params)
		.RunAndWait();
}


bool BrowserCommands::OnBeforeNavigation(const std::wstring& url, bool isMainFrame, int frameId)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = StringHelper::ToAString(url);
	params[1] = isMainFrame;
	params[2] = frameId;

	auto response = rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"beforeNavigation")
		.UseParameters(params)
		.RunAndWait();

	return response["result"].asBool();
}


void BrowserCommands::OnContextCreated(int64 frameId)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = frameId;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"contextCreated")
		.UseParameters(params)
		.RunAndWait();
}


void BrowserCommands::OnContextReleased(int64 frameId)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = frameId;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"contextReleased")
		.UseParameters(params)
		.Run();
}


void BrowserCommands::OnEditableNodeFocusChanged(bool isEditable, int64 frameId)
{
	PERFORMANCER;

	Json::Value params;
	params[0] = isEditable;
	params[1] = frameId;

	rpcProcessor.CreateRequest(GenerateMessagePrefix(), L"editableNodeFocusChanged")
		.UseParameters(params)
		.Run();
}


bool BrowserCommands::OnBeforeNavigationHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto url = StringHelper::ToWString(params[0].asString());
	auto isMainFrame = params[1].asBool();
	auto frameId = params[1].asInt();
	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	response["result"] = view->Events.OnBeforeNavigation(url, isMainFrame, frameId);
	return true;
}


bool BrowserCommands::OnContextCreatedHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	view->Events.OnContextCreated(params[0].asInt64());
	return true;
}


bool BrowserCommands::OnContextReleasedHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	view->Events.OnContextReleased(params[0].asInt64());
	return true;
}


bool BrowserCommands::OnViewIsReadyHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	view->Events.OnViewIsReady();

	return true;
}


bool BrowserCommands::OnChangeCurrentRenderProcess(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;
	auto sender = StringHelper::ToWString(params[0].asString());
	rpcProcessor.CreateNewSendTarget(sender);
	return true;
}


bool BrowserCommands::OnRendererCreatedHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto browserId = params[0].asInt();
	BrowserClient::GetInstance().GetViewManager().OnRendererCreated(browserId);
	return true;
}


bool BrowserCommands::OnExecuteRPCHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto requestJson = StringHelper::ToWString(params[0].asString());

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	Json::Value request;
	Json::Reader reader;
	if (!reader.parse(StringHelper::ToAString(requestJson), request))
	{
		std::wcerr << L"ERROR while receiving sitekioskobjectmodel call. Invalid JSON" << requestJson << std::endl;
		return false;
	}

	auto requestId = request["id"].asInt();
	std::wstring method = StringHelper::ToWString(request["method"].asString());
	if (method == L"ui.view.current")
	{
		Json::Value responseObj;
		responseObj["id"] = requestId;
		responseObj["result"] = view->GetId();

		response["result"] = responseObj;
	}
	else
	{
		std::wstringstream stringStream;
		stringStream << WebAppRPCProcessor::ObjectModelMessagePrefix;
		stringStream << view->GetId();
		stringStream << L":";

		response["result"] = BrowserClient::GetInstance().GetRPCProcessor().CreateRequestFromJson(stringStream.str(), requestId, requestJson)
			.RunAndWait(rpcProcessor.GetExitEventHandle());
	}

	return true;
}


bool BrowserCommands::OnPongHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	view->Pinger.Pong();
	return true;
}


void BrowserCommands::PerformPing(CefRefPtr<CefBrowser> browser)
{
	PERFORMANCER;

	if (browser == nullptr)// || browser->GetHost() == nullptr)
	{
		Pong();
		return;
	}

	CefRefPtr<CefV8Value> retval;
	CefRefPtr<CefV8Exception> exception;
	int frameCount = browser->GetFrameCount();
	std::vector<int64> frameIds;
	browser->GetFrameIdentifiers(frameIds);
	for (auto &frameId : frameIds)
	{
		auto frame = browser->GetFrame(frameId);
		if (frame == nullptr)
			continue;
		auto context = frame->GetV8Context();
		if (context == nullptr)
			continue;
		context->Eval("(function () { return true; })();", retval, exception); // this should hang until the V8 context is able to do something
	}

	Pong();
}


bool BrowserCommands::OnPingHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	if (!CefCurrentlyOn(TID_RENDERER))
	{
		CefPostTask(TID_RENDERER, base::Bind(&BrowserCommands::PerformPing, this, browser));
		return true;
	}
	PerformPing(browser);
	return true;
}


bool BrowserCommands::OnOpenDevToolsHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	CefPoint point;
	CefWindowInfo windowInfo;
	CefBrowserSettings settings;
	CefRefPtr<DevToolsWebClient> client = new DevToolsWebClient();
	auto host = browser->GetHost();
	windowInfo.SetAsPopup(host->GetWindowHandle(), "DevTools");
	host->ShowDevTools(windowInfo, client, settings, point);

	return true;
}


bool BrowserCommands::OnMustApplyObjectModelHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto url = StringHelper::ToWString(params[0].asString());
	response["result"] = BrowserClient::GetInstance().GetRPCProcessor().GetAppCommands().MustApplyObjectModel(url, rpcProcessor.GetExitEventHandle());
	return true;
}


bool BrowserCommands::OnEditableNodeFocusChangedHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto isEditable = params[0].asBool();
	auto frameId = params[1].asInt64();

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view != nullptr)
		view->Events.OnEditableNodeFocusChangedHandler(isEditable, frameId);

	return true;
}


bool BrowserCommands::OnPrintHandler(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto view = BrowserClient::GetInstance().GetViewManager().GetView(browser);
	if (view == nullptr)
		return false;

	if (view->Interaction.GetAllowPrinting())
		browser->GetHost()->Print();

	return true;
}


bool BrowserCommands::OnGetAssetsPath(const Json::Value& params, Json::Value& response)
{
	PERFORMANCER;

	auto assetsPath = CommandLine::GetInstance().GetStringArgument(L"assetsPath");

	response["result"] = StringHelper::ToAString(assetsPath);

	return true;
}



std::wstring BrowserCommands::GenerateMessagePrefix() const
{
	PERFORMANCER;

	int instanceId = this->rpcProcessor.GetIsBrowser() ? 0 : GetCurrentProcessId();
	return BrowserRPCProcessor::CommandPrefix + std::to_wstring(instanceId) + L":";
}
