#include "MessageProcessor.h"
#include "../../Utils/Profiler.h"
#include "Windows.h"

#include <iostream>
#include <sstream>


MessageProcessor::MessageProcessor(const std::wstring& sendTarget)
	: nextMessageHandlerId(0)
	, sendTarget(sendTarget)
	, isShuttingDown(false)
{

}


MessageProcessor::~MessageProcessor()
{
	Shutdown();
}


void MessageProcessor::OnMessageReceived(const std::wstring& message)
{
	PERFORMANCER;
	//std::unique_lock<std::mutex> lock(shutDownMutex);

	if (isShuttingDown)
		return;

	std::map<int, std::shared_ptr<Handler>> handlers;
	{
		std::unique_lock<std::mutex> lock(mutex);
		handlers = messageHandlers;
	}

	for (auto &handler : handlers)
	{
		if (handler.second->TryExecute(message))
			break;
	}

	{
		std::unique_lock<std::mutex> lock(mutex);
		for (auto iter = messageHandlers.begin(); iter != messageHandlers.end();)
		{
			if (iter->second->HasBeenUsed)
				messageHandlers.erase(iter++);
			else
				++iter;
		}
	}
}


void MessageProcessor::RemoveMessageHandler(int handlerId)
{
	std::unique_lock<std::mutex> lock(mutex);

	auto findResult = messageHandlers.find(handlerId);
	if (findResult == messageHandlers.end())
		return;
	findResult->second->Abort();
	messageHandlers.erase(findResult);
}


int MessageProcessor::AddMessageHandler(std::function<bool(const std::wstring&)> method, bool oneTimeUseOnly)
{
	std::unique_lock<std::mutex> lock(mutex);
	std::shared_ptr<MessageProcessor::Handler> handler(new MessageProcessor::Handler(method, oneTimeUseOnly));

	messageHandlers.insert(std::make_pair(nextMessageHandlerId, handler));
	return nextMessageHandlerId++;
}


int MessageProcessor::AddMessageHandler(const std::wstring& prefix, std::function<bool(const std::wstring&)> method, bool oneTimeUseOnly)
{
	return AddMessageHandler([prefix, method](const std::wstring& message) -> bool
	{
		if (message.compare(0, prefix.size(), prefix) == 0)
		{
			auto messageContent = message.substr(prefix.size());
			return method(messageContent);
		}
		return false;
	}, oneTimeUseOnly);
}


void MessageProcessor::Shutdown()
{
	std::unique_lock<std::mutex> lock(shutDownMutex);

	isShuttingDown = true;
}


void MessageProcessor::ChangeSendTarget(const std::wstring& newSendTarget)
{
	sendTarget = newSendTarget;
}


std::wstring MessageProcessor::GetSendTarget()
{
	return sendTarget;
}
