#include "NetPipeMessageProcessor.h"
#include "MessageProcessor.h"
#include "../Utils/ThreadHelper.h"

#include <iostream>
#include <fstream>
#include <string>

#include "windows.h"

NetPipeMessageProcessor::NetPipeMessageProcessor(const std::wstring& sendTarget, const std::wstring& receiverName)
	: MessageProcessor(sendTarget)
	, receiverName(receiverName)
{
}


NetPipeMessageProcessor::~NetPipeMessageProcessor()
{
	reader = nullptr;
	writers.clear();
}


void NetPipeMessageProcessor::Shutdown()
{
	reader = nullptr;
	writers.clear();
	MessageProcessor::Shutdown();
}


void NetPipeMessageProcessor::Start()
{
	std::unique_lock<std::mutex> lock(mutex);

	reader = std::unique_ptr<NetPipeReader>(new NetPipeReader(receiverName, std::bind(&NetPipeMessageProcessor::OnMessageReceived, this, std::placeholders::_1)));
	auto writer = std::shared_ptr<NetPipeWriter>(new NetPipeWriter(GetSendTarget(), std::bind(&NetPipeMessageProcessor::RemoveNetPipeWriter, this, std::placeholders::_1)));

	writers.push_back(writer);
}


void NetPipeMessageProcessor::Send(const std::wstring& message)
{
	std::unique_lock<std::mutex> lock(mutex);
	if (isShuttingDown)
		return;

	for (auto& writer : writers)
		writer->Send(message);
}


void NetPipeMessageProcessor::CreateNewSendTarget(const std::wstring& senderName)
{
	std::unique_lock<std::mutex> lock(mutex);

	MessageProcessor::ChangeSendTarget(senderName);

	for (auto& writer : writers)
		if (writer->GetName() == senderName)
		{
			writer->SetShutdownCallback(nullptr);
			auto iter = std::remove_if(writers.begin(), writers.end(), [&](const std::shared_ptr<NetPipeWriter> & otherWriter) -> bool
			{
				auto equal = otherWriter.get() == writer.get();
				return equal;
			});

			writers.erase(iter);
			
			break;
		}

	auto writer = std::shared_ptr<NetPipeWriter>(new NetPipeWriter(senderName, std::bind(&NetPipeMessageProcessor::RemoveNetPipeWriter, this, std::placeholders::_1)));
	writers.push_back(writer);
}


void NetPipeMessageProcessor::RemoveNetPipeWriter(NetPipeWriter* writer)
{
	std::unique_lock<std::mutex> lock(mutex);

	auto iter = std::remove_if(writers.begin(), writers.end(), [&](const std::shared_ptr<NetPipeWriter> & otherWriter) -> bool
	{
		auto equal = otherWriter.get() == writer;
		return equal;
	});

	writers.erase(iter);
}
