#include "SharedMemoryMessageProcessor.h"
#include "MessageProcessor.h"
#include "../Utils/ThreadHelper.h"

#include <iostream>
#include <fstream>
#include <string>

#include "windows.h"

SharedMemoryMessageProcessor::SharedMemoryMessageProcessor(const std::wstring& sendTarget, const std::wstring& receiverName)
	: MessageProcessor(sendTarget)
	, receiverName(receiverName)
{
}


SharedMemoryMessageProcessor::~SharedMemoryMessageProcessor()
{
	reader = nullptr;
	writers.clear();
}


void SharedMemoryMessageProcessor::Shutdown()
{
	reader = nullptr;
	writers.clear();
	MessageProcessor::Shutdown();
}


void SharedMemoryMessageProcessor::Start()
{
	std::unique_lock<std::mutex> lock(mutex);

	reader = std::unique_ptr<IPCReader>(new IPCReader(receiverName, std::bind(&SharedMemoryMessageProcessor::OnMessageReceived, this, std::placeholders::_1)));
	auto writer = std::shared_ptr<IPCWriter>(new IPCWriter(GetSendTarget(), std::bind(&SharedMemoryMessageProcessor::RemoveWriter, this, std::placeholders::_1)));

	writers.push_back(writer);
}


void SharedMemoryMessageProcessor::Send(const std::wstring& message)
{
	std::unique_lock<std::mutex> lock(mutex);
	if (isShuttingDown)
		return;

	for (auto& writer : writers)
		writer->Write(message);
}


void SharedMemoryMessageProcessor::CreateNewSendTarget(const std::wstring& senderName)
{
	std::unique_lock<std::mutex> lock(mutex);

	MessageProcessor::ChangeSendTarget(senderName);

	for (auto& writer : writers)
		if (writer->GetName() == senderName)
		{
			writer->SetShutdownCallback(nullptr);
			auto iter = std::remove_if(writers.begin(), writers.end(), [&](const std::shared_ptr<IPCWriter> & otherWriter) -> bool
			{
				auto equal = otherWriter.get() == writer.get();
				return equal;
			});

			writers.erase(iter);
			
			break;
		}

	auto writer = std::shared_ptr<IPCWriter>(new IPCWriter(senderName, std::bind(&SharedMemoryMessageProcessor::RemoveWriter, this, std::placeholders::_1)));
	writers.push_back(writer);
}


void SharedMemoryMessageProcessor::RemoveWriter(IPCWriter* writer)
{
	std::unique_lock<std::mutex> lock(mutex);

	auto iter = std::remove_if(writers.begin(), writers.end(), [&](const std::shared_ptr<IPCWriter> & otherWriter) -> bool
	{
		auto equal = otherWriter.get() == writer;
		return equal;
	});

	writers.erase(iter);
}
