#include "NetPipeReader.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/ThreadHelper.h"
#include "../../Utils/Profiler.h"
#include <sstream>


NetPipeReader::NetPipeReader(const std::wstring& name, const std::function<void(const std::wstring&)>& onMessage)
	: pipeName(L"\\\\.\\pipe\\" + name)
	, exitEvent(events.front())
{
	for (size_t i = 0; i < events.size(); ++i)
		events[i] = ManualResetEvent(false);
	exitEvent = events.front();

	// The initial loop creates several instances of a named pipe 
	// along with an event object for each instance.  An 
	// overlapped ConnectNamedPipe operation is started for 
	// each instance. 

	for (std::size_t i = 0; i < pipes.size(); i++)
		pipes[i].Connect(SERVER_PIPE_INSTANCES, pipeName, events[i + 1], onMessage);

	thread = std::unique_ptr<std::thread>(new std::thread(&NetPipeReader::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "NetPipe Reader " + StringHelper::ToAString(name));
}


NetPipeReader::~NetPipeReader()
{
	exitEvent.Set();

	CancelSynchronousIo(static_cast<HANDLE>(thread->native_handle()));

	// It can happen that we are still processing messages which may hold a reference to a View and thus to a BrowserRPCProcessor and thus to a Server instance
	// After the message has been processed those references get freed, in some cases these are the last references and will result in a call to the Server DTor
	// Which will make the current thread not joinable as the currentThread is the thread we want to join. This should be prevented through some other code (View::Close will shutdown the BroweserRPCProcessor)
	// But to be sure we just check this here as the bug was very rare and reproduction of it was very difficult even before the potential fix of the issue.
	if (thread->get_id() != std::this_thread::get_id())
		thread->join();
	else
		thread->detach();

	FreeResources();
}


void NetPipeReader::ProcessMessagesAsync()
{
	auto waitEventHandles = ResetEventBase::GenerateHandlesArray(events);
	while (true)
	{
		PERFORMANCER;

		// Wait for the event object to be signaled, indicating 
		// completion of an overlapped read, write, or 
		// connect operation. 
		auto waitResult = ResetEventBase::Wait(waitEventHandles);
		if (waitResult == WAIT_OBJECT_0) // exit
			break;

		// waitResult shows which pipe completed the operation. 
		int pipeIndex = (waitResult - WAIT_OBJECT_0);  // determines which pipe 
		pipeIndex--; // correct index because first Wait Object was our exit event
		
		// Get the result if the operation was pending. 
		pipes[pipeIndex].ProcessMessages();
	}

	FreeResources();
}


void NetPipeReader::FreeResources()
{
	for (std::size_t i = 0; i < pipes.size(); i++)
		pipes[i].Close();
}
