#include "NetPipeWriter.h"
#include "NetPipeReader.h" // SERVER_PIPE_BUFSIZE
#include "../../Utils/StringHelper.h"
#include "../../Utils/ThreadHelper.h"
#include "../../Utils/Profiler.h"


NetPipeWriter::NetPipeWriter(const std::wstring& name)
	: pipeName(L"\\\\.\\pipe\\" + name)
	, name(name)
	, shutdownCallback([](NetPipeWriter* writer){})
	, exitEvent(false)
	, hasMessagesEvent(false)
{
	thread = std::unique_ptr<std::thread>(new std::thread(&NetPipeWriter::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "NetPipe Writer " + StringHelper::ToAString(name));
}


NetPipeWriter::NetPipeWriter(const std::wstring& name, const std::function<void(NetPipeWriter*)>& shutdownCallback)
	: pipeName(L"\\\\.\\pipe\\" + name)
	, name(name)
	, shutdownCallback(shutdownCallback)
	, exitEvent(false)
	, hasMessagesEvent(false)
{
	thread = std::unique_ptr<std::thread>(new std::thread(&NetPipeWriter::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "NetPipe Writer " + StringHelper::ToAString(name));
}


NetPipeWriter::~NetPipeWriter()
{
	exitEvent.Set();
	if (thread->get_id() != std::this_thread::get_id())
		thread->join();
	else
		thread->detach();
}


void NetPipeWriter::SetShutdownCallback(const std::function<void(NetPipeWriter*)>& shutdownCallback)
{
	this->shutdownCallback = shutdownCallback;
}


Handle NetPipeWriter::EstablishConnection()
{
	PERFORMANCER;

	Handle pipeInstance;

	// Try to open a named pipe; wait for it, if necessary. 
	while (exitEvent.Wait(0) != 0)
	{
		pipeInstance = CreateFile(
			pipeName.c_str(),   // pipe name 
			GENERIC_READ |  // read and write access 
			GENERIC_WRITE,
			0,              // no sharing 
			NULL,           // default security attributes
			OPEN_EXISTING,  // opens existing pipe 
			0,              // default attributes 
			NULL);

		if (pipeInstance)
			break;

		auto lastError = GetLastError();
		if (lastError != ERROR_PIPE_BUSY)
		{
			Sleep(1);
			continue;
		}

		if (!WaitNamedPipe(pipeName.c_str(), 1000))
			continue;
	}

	return pipeInstance;
}


bool NetPipeWriter::SendMessageOverPipe(const Handle& pipeInstance, const std::string& message)
{
	PERFORMANCER;

	DWORD bytesWritten;

	const int maxMessageSize = SERVER_PIPE_BUFSIZE - 4;
	int messageSize = (int)message.size();
	bool isFirstPart = true;
	
	auto messageData = message.c_str();
	int currentMessageDataPos = 0;
	while (currentMessageDataPos != messageSize)
	{
		int currentPartSize = min(maxMessageSize, messageSize - currentMessageDataPos);
		int currentBufPos = 0;
		char* buffer;
		if (isFirstPart)
		{
			buffer = new char[currentPartSize + 4];

			buffer[0] = ((char*)(&messageSize))[0];
			buffer[1] = ((char*)(&messageSize))[1];
			buffer[2] = ((char*)(&messageSize))[2];
			buffer[3] = ((char*)(&messageSize))[3];
			currentBufPos = 4;
			isFirstPart = false;
		}
		else
		{
			buffer = new char[currentPartSize];
		}
		memcpy_s(&buffer[currentBufPos], maxMessageSize, &messageData[currentMessageDataPos], currentPartSize);
		currentMessageDataPos += currentPartSize;

		auto success = WriteFile(
			pipeInstance,                  // pipe handle 
			buffer,             // message 
			currentPartSize + currentBufPos,              // message length 
			&bytesWritten,             // bytes written 
			NULL);                  // not overlapped 

		/*if (success == FALSE)
			StringHelper::DebugPrint(L"WARNING ({0}): Unable to send message: {1}", ::GetLastError(), StringHelper::ToWString(message));*/

		delete[] buffer;

		if (success == FALSE)
			return false;
	}

	return true;
}


void NetPipeWriter::ProcessMessagesAsync()
{
	PERFORMANCER;

	auto pipeInstance = EstablishConnection();
	if (!pipeInstance)
		return;

	//DWORD bytesWritten;
	DWORD messageReadMode = PIPE_READMODE_MESSAGE;
	HANDLE waitEvents[2] = { exitEvent, hasMessagesEvent };
	while (true)
	{
		auto result = WaitForMultipleObjects(2, waitEvents, false, INFINITE);
		if (result == WAIT_OBJECT_0)
			break;
		else if (result == WAIT_OBJECT_0 + 1)
		{
			// The pipe connected; change to message-read mode. 
			if (SetNamedPipeHandleState(
				pipeInstance,    // pipe handle 
				&messageReadMode,  // new pipe mode 
				NULL,     // don't set maximum bytes 
				NULL) == FALSE)
			{
				break;
			}

			std::unique_lock<std::mutex> lock(mutex);
			bool sendFailed = false;
			while (!pendingMessages.empty())
			{
				std::wstring message = pendingMessages.front();
				std::string messageA = StringHelper::ToAString(message);

				SendMessageOverPipe(pipeInstance, messageA);

				pendingMessages.pop();

			}
		}
	}
	if (shutdownCallback)
		shutdownCallback(this);
}


void NetPipeWriter::Send(const std::wstring& message)
{
	std::unique_lock<std::mutex> lock(mutex);
	pendingMessages.push(message);
	hasMessagesEvent.Set();
}
