#include "PipeInstance.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/Profiler.h"


PipeInstance::PipeInstance()
	: readBytes(0)
	, bytesToWrite(0)
	, state(PipeState::Connecting)
	, hasPendingIO(false)
	, messageLength(0)
	, receivedMessageLength(0)
{
	ZeroMemory(&overlapped, sizeof(OVERLAPPED));
}

PipeInstance::~PipeInstance()
{
	Close();
}


void PipeInstance::Close()
{
	if (pipeInstance)
	{
		DisconnectNamedPipe(pipeInstance);
	}
}

void PipeInstance::Connect(int pipeInstanceCount, const std::wstring& pipeName, const Handle& waitEvent, const std::function<void(const std::wstring&)>& onMessageCallback)
{
	onMessage = onMessageCallback;
	overlapped.hEvent = waitEvent;

	pipeInstance = CreateNamedPipe(
		pipeName.c_str(),            // pipe name 
		PIPE_ACCESS_DUPLEX |     // read/write access 
		FILE_FLAG_OVERLAPPED,    // overlapped mode 
		PIPE_TYPE_MESSAGE |      // message-type pipe 
		PIPE_READMODE_MESSAGE |  // message-read mode 
		PIPE_WAIT,               // blocking mode 
		pipeInstanceCount,               // number of instances 
		SERVER_PIPE_BUFSIZE * sizeof(char),   // output buffer size 
		SERVER_PIPE_BUFSIZE * sizeof(char),   // input buffer size 
		PIPE_TIMEOUT,            // client time-out 
		NULL);                   // default security attributes 

	if (!pipeInstance)
	{
		// ERROR HANDLING?
		return;
	}
	
	ConnectToNewClient();
	state = hasPendingIO ? PipeState::Connecting : PipeState::Reading;
}


void PipeInstance::DisconnectAndReconnect()
{
	DisconnectNamedPipe(pipeInstance);
	ConnectToNewClient();
	state = hasPendingIO ? PipeState::Connecting : PipeState::Reading;
}


void PipeInstance::ConnectToNewClient()
{
	hasPendingIO = false;
	overlapped.Offset = 0;
	overlapped.OffsetHigh = 0;

	// Start an overlapped connection for this pipe instance. 
	auto isConnected = ConnectNamedPipe(pipeInstance, &overlapped);

	// Overlapped ConnectNamedPipe should return zero. 
	if (isConnected)
		return;

	switch (GetLastError())
	{
		// The overlapped connection in progress. 
		case ERROR_IO_PENDING:
			hasPendingIO = true;
			break;

		// Client is already connected, so signal an event. 
		case ERROR_PIPE_CONNECTED:
			if (SetEvent(overlapped.hEvent))
				break;

		// If an error occurs during the connect operation... 
		default:
			hasPendingIO = false;
	}
}

bool PipeInstance::TryGetPendingIOResult()
{
	PERFORMANCER;

	if (!hasPendingIO)
		return true;

	DWORD bytesTransfered;
	bool wasSuccess = GetOverlappedResult(
		pipeInstance, // handle to pipe 
		&overlapped, // OVERLAPPED structure 
		&bytesTransfered,            // bytes transferred 
		FALSE) == TRUE;            // do not wait 

	switch (state)
	{
		// Pending connect operation 
		case PipeState::Connecting:
			if (!wasSuccess)
				return false;
			state = PipeState::Reading;
			break;

		// Pending read operation 
		case PipeState::Reading:
			if (!wasSuccess || bytesTransfered == 0)
			{
				auto lastError = GetLastError();
				DisconnectAndReconnect();
				return false;
			}
			readBytes = bytesTransfered;
			state = PipeState::Writing;
			break;

		// Pending write operation 
		case PipeState::Writing:
			if (!wasSuccess || bytesTransfered != bytesToWrite)
			{
				DisconnectAndReconnect();
				return false;
			}
			state = PipeState::Reading;
			break;

		default:
			return false;
	}

	return true;
}

void PipeInstance::ProcessMessages()
{
	PERFORMANCER;

	if (!TryGetPendingIOResult())
		return;

	switch (state)
	{
		// READING_STATE: 
		// The pipe instance is connected to the client 
		// and is ready to read a request from the client. 
		case PipeState::Reading:
		{
			bool fSuccess = ReadFile(
				pipeInstance,
				requestBuffer.data(),
				requestBuffer.size() * sizeof(char),
				&readBytes,
				&overlapped) == TRUE;

			// The read operation completed successfully. 
			if (fSuccess && readBytes != 0)
			{
				hasPendingIO = true;
				return;
			}

			// The read operation is still pending. 
			auto dwErr = GetLastError();
			if (!fSuccess && (dwErr == ERROR_IO_PENDING))
			{
				hasPendingIO = true;
				return;
			}

			// An error occurred; disconnect from the client. 
			if (dwErr != 109) // pipes closed
				OutputDebugString((L"NetPipe has a problem, reconnecting. Error was: " + std::to_wstring(dwErr) + L"\n").c_str());

			DisconnectAndReconnect();
			break;
		}


		// WRITING_STATE: 
		// The request was successfully read from the client. 
		// Get the reply data and write it to the client. 
		case PipeState::Writing:
		{
			std::string messagePart;

			if (messageLength == 0)
			{
				messagePart = { requestBuffer.data() + 4, readBytes - 4 };
				messageLength = *((int*)requestBuffer.data());
				receivedMessageLength += readBytes - 4;
			}
			else
			{
				messagePart = { requestBuffer.data(), readBytes };
				receivedMessageLength += readBytes;
			}
			receivedMessage << messagePart;

			if (receivedMessageLength == messageLength)
			{
				auto message = StringHelper::ToWString(receivedMessage.str());
				receivedMessageLength = 0;
				messageLength = 0;
				receivedMessage.str(""); // clears the stream
				receivedMessage.clear(); // clears the fail flags

				if (!message.empty())
					onMessage(message);
			}

			hasPendingIO = false;
			state = PipeState::Reading;

			break;
		}
	}
}