#include "RPCProcessor.h"
#include "MessageProcessor/MessageProcessor.h"
#include "MessageProcessor/StdInOutMessageProcessor.h"
#include "../Browser/BrowserClient.h"
#include "../Thread/ManualResetEvent.h"
#include "../Utils/StringHelper.h"
#include "../Utils/Profiler.h"
#include "include/cef_app.h"

#include <fstream>
#include <iostream>
#include <json\json.h>



///////////////////////////////////////////////////////////////////////////
RPCProcessor::RPCProcessor(MessageLoggingLevel logLevel)
	: requestCount(0)
	, logLevel(logLevel)
	, exitEvent(false)
{
#ifndef DEBUG
	std::ifstream debugTxtFile(".\\debug.txt");
	if (!debugTxtFile)
		logLevel = MessageLoggingLevel::None;
#endif
}

RPCProcessor::~RPCProcessor()
{
	Shutdown();
}


void RPCProcessor::Shutdown()
{
	exitEvent.Set();
}

std::wstring RPCProcessor::GenerateRequestString(bool useId, const RPCRequestInternal& request, Json::Value* outRequest)
{
	PERFORMANCER;

	auto messagePrefix = request.GetPrefix();
	auto requestString = request.GetRawJsonRequest();
	if (requestString.empty())
	{
		auto method = request.GetMethod();
		auto params = request.GetParameters();

		Json::Value requestJson;
		if (useId)
		{
			auto requestId = request.GetRequestId();
			requestJson["id"] = std::to_string(requestId);
		}
		requestJson["method"] = StringHelper::ToAString(method);
		requestJson["params"] = params;

		if (outRequest != nullptr)
			*outRequest = requestJson;

		Json::FastWriter writer;
		auto jsonStringA = writer.write(requestJson);
		jsonStringA.pop_back();
		requestString = messagePrefix + StringHelper::ToWString(jsonStringA);
		return requestString;
	}
	else
	{
		if (outRequest != nullptr)
		{
			Json::Reader reader;
			reader.parse(StringHelper::ToAString(requestString), *outRequest);
		}

		return messagePrefix + requestString;
	}
}


RPCRequest RPCProcessor::CreateRequest(const std::wstring& prefix, const std::wstring& method, int requestId)
{
	return RPCRequest(*this, prefix, method, requestId);
}


RPCRequest RPCProcessor::CreateRequestFromJson(const std::wstring& prefix, int requestId, const std::wstring& jsonString)
{
	return RPCRequest(*this, prefix, requestId, jsonString);
}


RPCProcessor::Response RPCProcessor::SendRequest(const RPCRequest& request)
{
	PERFORMANCER;

	auto internalRequest = RPCRequestInternal(request);
	auto requestId = internalRequest.GetRequestId();
	if (requestId == -1)
		internalRequest.SetRequestId(requestCount++);

	if (internalRequest.HasResponseCallback())
		return RunAsyncRequest(internalRequest);
	else if (internalRequest.IsWaitingForResponse())
		return RunRequestAndWaitForResult(internalRequest);
	else
		return RunRequestAndIgnoreResult(internalRequest);
}


RPCProcessor::Response RPCProcessor::RunAsyncRequest(RPCRequestInternal& request)
{
	PERFORMANCER;

	Json::Value requestJson;
	auto requestString = GenerateRequestString(true, request, &requestJson);
	auto requestId = request.GetRequestId();
	auto messagePrefix = request.GetPrefix();

	auto cancelable = std::shared_ptr<Cancelable>(new Cancelable(requestId));
	auto responseCallback = request.GetResponseCallback();
	request.RemoveResponseCallback();

	int handlerId = GetMessageProcessor().AddMessageHandler(messagePrefix, [this, messagePrefix, requestJson, responseCallback, requestId, cancelable](const std::wstring& message) -> bool
	{
		Json::Reader reader;
		Json::Value response;
		if (!reader.parse(StringHelper::ToAString(message), response) || response["id"].empty())
			return false;
		if (response["id"].asString() != std::to_string(requestId))
			return false;

		cancelable->Done();
		responseCallback(response);

		LogMessageTransfer(messagePrefix, requestJson, response, false);

		return true;
	}, true);


	GetMessageProcessor().Send(requestString);

	cancelable->OnCancel([&, handlerId, messagePrefix, requestJson]() -> void {
		LogMessageTransfer(messagePrefix, requestJson, Json::Value::null, false);
		GetMessageProcessor().RemoveMessageHandler(handlerId);
	}, true);

	return Response(cancelable);
}


RPCProcessor::Response RPCProcessor::RunRequestAndWaitForResult(RPCRequestInternal& request)
{
	PERFORMANCER;

	Json::Value requestJson;
	auto requestString = GenerateRequestString(true, request, &requestJson);
	auto responsePrefix = request.GetResponsePrefix();
	auto requestId = request.GetRequestId();
	auto messagePrefix = request.GetPrefix();

	ManualResetEvent waitHandle(false);
	Json::Value responseJson;

	auto responseHandler = [this, messagePrefix, requestJson, &responseJson, requestId, &waitHandle](const std::wstring& message) -> bool
	{
		PERFORMANCER;

		Json::Reader reader;
		Json::Value tmpResponse;
		if (!reader.parse(StringHelper::ToAString(message), tmpResponse))
			return false;
		auto method = tmpResponse["method"].asString();
		if (!method.empty() || tmpResponse["id"].empty())
			return false;
		if (tmpResponse["id"].asString() != std::to_string(requestId))
			return false;

		responseJson = tmpResponse;
		waitHandle.Set();
		LogMessageTransfer(messagePrefix, requestJson, tmpResponse, false);

		return true;
	};

	auto messageHandlerId = GetMessageProcessor().AddMessageHandler(responsePrefix.empty() ? messagePrefix : responsePrefix, responseHandler, true);
	GetMessageProcessor().Send(requestString);

	auto abortHandle = request.GetAbortHandle();
	DWORD result;
	if (abortHandle != nullptr)
	{
		HANDLE waitEvents[3] = { exitEvent, waitHandle, abortHandle };
		result = WaitForMultipleObjects(3, waitEvents, false, 10 * 1000);
	}
	else
	{
		HANDLE waitEvents[2] = { exitEvent, waitHandle };
		result = WaitForMultipleObjects(2, waitEvents, false, 10 * 1000);
	}

	if (result != WAIT_OBJECT_0 + 1)
	{
		LogMessageTransfer(messagePrefix, requestJson, Json::Value::null, false);
		GetMessageProcessor().RemoveMessageHandler(messageHandlerId);
		Json::Value errorResponse;
		errorResponse["id"] = requestId;
		errorResponse["error"] = "Error while executing RPC: Connection lost.";
		return Response(errorResponse);
	}
	return Response(responseJson);
}


RPCProcessor::Response RPCProcessor::RunRequestAndIgnoreResult(RPCRequestInternal& request)
{
	PERFORMANCER;

	Json::Value requestJson;
	auto requestString = GenerateRequestString(false, request, &requestJson);
	auto messagePrefix = request.GetPrefix();

	GetMessageProcessor().Send(requestString);
	LogMessageTransfer(messagePrefix, requestJson, Json::Value::null, false);
	return Response(Json::Value::null);
}


bool RPCProcessor::HandleJsonRequest(const JsonHandlerInfo& handlerInfo, const std::wstring& message, int subTargetId)
{
	PERFORMANCER;

	Json::Value request;
	Json::Value response;
	Json::Reader reader;
	if (!reader.parse(StringHelper::ToAString(message), request))
		return false;

	auto wantsResponse = !request["id"].empty();
	if (wantsResponse)
		response["id"] = request["id"].asString();

	auto method = StringHelper::ToWString(request["method"].asString());
	if (method.empty())
		return false;

	auto params = request["params"];
	auto responsePrefix = handlerInfo.ResponsePrefix;
	for (auto &handler : handlerInfo.Handlers)
	{
		if (handler.Method == method && TryHandleRequest(handler, request, responsePrefix, params, response, subTargetId))
			return true;
	}
	if (wantsResponse)
	{
		Json::FastWriter writer;
		auto requestJsonString = writer.write(request);
		requestJsonString.pop_back();
		response["error"] = "No handler was found for RPC request: " + requestJsonString;

		SendResponse(responsePrefix, response);
	}
	return true;
}


bool RPCProcessor::TryHandleRequest(const RPCProcessor::Handler& handler, const Json::Value& request, const std::wstring& responsePrefix, const Json::Value& params, Json::Value& response, int subTargetId)
{
	PERFORMANCER;

	auto wantsResponse = !request["id"].empty();
	try
	{
		if (!handler.IsBlockingCall)
		{
			auto couldHandle = handler.FunctionNonBlocking(params, response, [&, responsePrefix, request](const Json::Value& deferredResponse) -> void
			{
				LogMessageTransfer(responsePrefix, request, deferredResponse, true);
				SendResponse(responsePrefix, deferredResponse);
			}, subTargetId);
			if (couldHandle)
				return true;
		}
		else
		{
			if (handler.Function(params, response, subTargetId))
			{
				LogMessageTransfer(responsePrefix, request, response, true);
				if (wantsResponse)
					SendResponse(responsePrefix, response);
				return true;
			}
		}

	}
	catch (...)
	{
		Json::FastWriter writer;
		auto requestJsonString = writer.write(request);
		requestJsonString.pop_back();

		if (!wantsResponse)
		{
			OutputDebugString((L"Error while executing RPC: " + StringHelper::ToWString(requestJsonString) + L"\n").c_str());
			return true;
		}

		response["id"] = request["id"].asString();
		response["error"] = "Error while executing RPC: " + requestJsonString;
		SendResponse(responsePrefix, response);
		return true;
	}

	return false;
}


void RPCProcessor::SendResponse(const std::wstring& messagePrefix, const Json::Value& response)
{
	PERFORMANCER;

	Json::FastWriter writer;
	auto jsonString = writer.write(response);
	jsonString.pop_back();

	GetMessageProcessor().Send(messagePrefix + StringHelper::ToWString(jsonString));
}


void RPCProcessor::LogMessageTransfer(const std::wstring& prefix, const Json::Value& request, const Json::Value& response, bool receivedRequest)
{
	PERFORMANCER;

	if (logLevel == MessageLoggingLevel::None)
		return;

	std::wstringstream ss;
	Json::FastWriter writer;
	auto method = request["method"].asString();

	if (method == "checkMustCancelRequest") // dont log this spammy but important message
		return;

	auto paramsJson = request["params"];
	auto params = paramsJson.empty() ? "" : writer.write(paramsJson);
	if (!params.empty())
	{
		params.pop_back();
		params.pop_back();
		params.erase(0, 1);

		if (params.size() > 200)
		{
			params.resize(200);
			params += " ...";
		}
	}
	auto resultJson = response["result"];
	auto result = resultJson.empty() ? "" : writer.write(resultJson);
	if (!result.empty())
		result.pop_back();

	ss << prefix
	   << std::wstring(receivedRequest ? L" received: " : L" send: ") 
	   << StringHelper::ToWStringNoException(method) 
	   << L"(" << StringHelper::ToWStringNoException(params) 
	   << L")";

	if (!result.empty())
		ss << L" => " << StringHelper::ToWStringNoException(result) << std::endl;
	else
		ss << std::endl;
	
	OutputDebugString(ss.str().c_str());
}