#pragma once

#include "MessageProcessor/MessageProcessor.h"
#include "Cancelable.h"
#include "RPCRequest.h"
#include "../Thread/ManualResetEvent.h"

#include <string>
#include <vector>
#include <atomic>
#include <mutex>
#include <json/json.h>

enum MessageLoggingLevel
{
	None,
	ResponseAndRequests
};

class RPCProcessor
{
public:
	typedef std::function<bool(const Json::Value& parameters, Json::Value response, const std::function<void(const Json::Value&)>& onFinished, int subTargetId)> RPCFunctionNonBlocking;
	typedef std::function<bool(const Json::Value& parameters, Json::Value& response, int subTargetId)> RPCFunction;

	class Handler
	{
	public:
		Handler(const std::wstring method, RPCFunction function)
			: Method(method)
			, Function(function)
			, IsBlockingCall(true)
		{
		}

		Handler(const std::wstring method, RPCFunctionNonBlocking function)
			: Method(method)
			, FunctionNonBlocking(function)
			, IsBlockingCall(false)
		{
		}

		std::wstring Method;
		bool IsBlockingCall;
		RPCFunction Function;
		RPCFunctionNonBlocking FunctionNonBlocking;
	};

	struct JsonHandlerInfo
	{
		std::vector<Handler> Handlers;
		std::wstring ResponsePrefix;
	};

	class Response
	{
	public:
		Response(std::shared_ptr<Cancelable> cancelable) : cancelable(cancelable) {}
		Response(const Json::Value& response) : response(response) {}

		Json::Value GetJsonResponse()
		{
			return response;
		}

		std::shared_ptr<Cancelable> GetCancelable()
		{
			return cancelable;
		}

	private:
		Json::Value response;
		std::shared_ptr<Cancelable> cancelable;
	};


public:
	RPCProcessor(MessageLoggingLevel logLevel = MessageLoggingLevel::None);
	~RPCProcessor();

	friend class RPCRequest;

	RPCRequest CreateRequest(const std::wstring& prefix, const std::wstring& method, int requestId = -1);
	RPCRequest CreateRequestFromJson(const std::wstring& prefix, int requestId, const std::wstring& jsonString);
	HANDLE GetExitEventHandle() { return (HANDLE)exitEvent; }

protected:
	Response SendRequest(const RPCRequest& request);

	std::wstring GenerateRequestString(bool useId, const RPCRequestInternal& request, Json::Value* outRequest = nullptr);
	Response RunAsyncRequest(RPCRequestInternal& request);
	Response RunRequestAndWaitForResult(RPCRequestInternal& request);
	Response RunRequestAndIgnoreResult(RPCRequestInternal& request);

protected:
	void Shutdown();
	virtual MessageProcessor& GetMessageProcessor() = 0;
	bool HandleJsonRequest(const JsonHandlerInfo& handlerInfo, const std::wstring& message, int subTargetId = -1);
	bool TryHandleRequest(const RPCProcessor::Handler& handler, const Json::Value& request, const std::wstring& responsePrefix, const Json::Value& params, Json::Value& response, int subTargetId);
	void SendResponse(const std::wstring& messagePrefix, const Json::Value& response);

protected:
	ManualResetEvent exitEvent;
	std::mutex mutex;
	std::atomic<int> requestCount;

private:
	void LogMessageTransfer(const std::wstring& prefix, const Json::Value& request, const Json::Value& response, bool receivedRequest);
	MessageLoggingLevel logLevel;
};
