#include "IPCReader.h"
#include <sstream>
#include "../../Utils/StringHelper.h"
#include "../../Utils/ThreadHelper.h"
#include "../../Utils/Profiler.h"





IPCReader::IPCReader(const std::wstring& name, const std::function<void(const std::wstring&)>& onMessage, size_t capacity)
	: dataMutex(false, L"mutex://" + name)
	, dataAvailableEvent(false, L"dataAvailableEvent://" + name)
	, exitEvent(false)
	, memoryMappedFile(L"memory://" + name, capacity + 8)
	, capacity(capacity)
	, name(name)
	, onMessage(onMessage)
	, messageLength(0)
	, receivedMessageLength(0)
	, readBuffer(new char[capacity])
{
	thread = std::unique_ptr<std::thread>(new std::thread(&IPCReader::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "IPC Reader " + StringHelper::ToAString(name));
}


IPCReader::IPCReader()
{
}


IPCReader::~IPCReader()
{
	exitEvent.Set();

	// It can happen that we are still processing messages which may hold a reference to a View and thus to a BrowserRPCProcessor and thus to a Server instance
	// After the message has been processed those references get freed, in some cases these are the last references and will result in a call to the Server DTor
	// Which will make the current thread not joinable as the currentThread is the thread we want to join. This should be prevented through some other code (View::Close will shutdown the BroweserRPCProcessor)
	// But to be sure we just check this here as the bug was very rare and reproduction of it was very difficult even before the potential fix of the issue.
	if (thread->get_id() != std::this_thread::get_id())
		thread->join();
	else
		thread->detach();

	delete[] readBuffer;
}


size_t IPCReader::Read()
{
	size_t readBytes = 0;
	dataAvailableEvent.Wait();

	dataMutex.Wait();
	
	{
		auto fileMap = memoryMappedFile.CreateViewAccessor();
		readBytes = *(size_t*)(fileMap.Data() + 0);
		memcpy_s(readBuffer, readBytes, fileMap.Data() + 8, readBytes);
		*(size_t*)(fileMap.Data() + 0) = 0;
	}

	dataAvailableEvent.Reset();
	dataMutex.ReleaseMutex();

	return readBytes;
}


void IPCReader::ProcessMessagesAsync()
{
	std::array<HANDLE, 2> waitEventHandles;
	waitEventHandles[0] = exitEvent;
	waitEventHandles[1] = dataAvailableEvent;
	while (true)
	{
		PERFORMANCER;

		// Wait for the event object to be signaled, indicating 
		// completion of an overlapped read, write, or 
		// connect operation. 
		auto waitResult = ResetEventBase::Wait(waitEventHandles);
		if (waitResult == WAIT_OBJECT_0) // exit
			break;

		auto readBytes = Read();

		std::string messagePart;

		if (messageLength == 0)
		{
			messageLength = *((int*)readBuffer);
			receivedMessageLength += readBytes - 4;
			receivedMessage << std::string{ readBuffer + 4, readBytes - 4 };
		}
		else
		{
			receivedMessageLength += readBytes;
			receivedMessage << std::string{ readBuffer, readBytes };
		}

		if (receivedMessageLength == messageLength)
		{
			auto message = StringHelper::ToWString(receivedMessage.str());
			receivedMessageLength = 0;
			messageLength = 0;
			receivedMessage.str(""); // clears the stream
			receivedMessage.clear(); // clears the fail flags

			if (!message.empty())
				onMessage(message);
		}
	}
}
