#include "IPCWriter.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/ThreadHelper.h"
#include "../../Utils/Profiler.h"

bool IPCWriter::CanOpen(const std::wstring& name)
{
	auto memoryMappedFile = MemoryMappedFile(L"memory://" + name);
	return memoryMappedFile.IsConnected();
}


bool IPCWriter::CanCreate(const std::wstring& name)
{
	return !IPCWriter::CanOpen(name);
}


IPCWriter::IPCWriter(const std::wstring& name, const std::function<void(IPCWriter*)>& shutdownCallback, size_t capacity)
	: name(name)
	, shutdownCallback(shutdownCallback)
	, capacity(capacity)
	, exitEvent(false)
	, hasMessagesEvent(false)
{
	thread = std::unique_ptr<std::thread>(new std::thread(&IPCWriter::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "IPC Writer " + StringHelper::ToAString(name));
}


IPCWriter::~IPCWriter()
{
	exitEvent.Set();
	if (thread->get_id() != std::this_thread::get_id())
		thread->join();
	else
		thread->detach();
}


bool IPCWriter::TryConnect(const std::wstring& name)
{
	memoryMappedFile = MemoryMappedFile(L"memory://" + name);
	if (!memoryMappedFile.IsConnected())
		return false;
	dataMutex = Mutex(false, L"mutex://" + name);
	dataAvailableEvent = ResetEventBase(L"dataAvailableEvent://" + name);

	return true;
}


void IPCWriter::SetShutdownCallback(const std::function<void(IPCWriter*)>& shutdownCallback)
{
	this->shutdownCallback = shutdownCallback;
}


void IPCWriter::Write(const char* buffer, size_t size)
{
	size_t offset = 0;
	while (offset != size)
	{
		offset += WriteInternal(buffer, size, offset);
	}
}


size_t IPCWriter::WriteInternal(const char* buffer, size_t size, size_t offset)
{
	size_t bytesToWrite = 0;
	dataMutex.Wait();

	{
		auto fileMap = memoryMappedFile.CreateViewAccessor();

		size_t position = *(size_t*)(fileMap.Data() + 0);
		size_t remainingCapacity = capacity - position;

		bytesToWrite = min(remainingCapacity, size - offset);
		if (bytesToWrite == 0)
		{
			dataMutex.ReleaseMutex();
			return bytesToWrite;
		}

		/*StringHelper::DebugPrint(L"NOW!");
		Sleep(5000);*/
		memcpy_s(fileMap.Data() + 8 + position, capacity, buffer + offset, bytesToWrite);
		*(size_t*)(fileMap.Data() + 0) = position + bytesToWrite;
	}
	dataAvailableEvent.Set();
	dataMutex.ReleaseMutex();
	return bytesToWrite;
}


void IPCWriter::Write(const std::wstring& message)
{
	std::unique_lock<std::mutex> lock(mutex);
	pendingMessages.push(message);
	hasMessagesEvent.Set();
}


void IPCWriter::ProcessMessagesAsync()
{
	PERFORMANCER;

	while (true)
	{
		if (exitEvent.Wait(0) == 0)
			return;

		if (TryConnect(name))
			break;
	}

	DWORD messageReadMode = PIPE_READMODE_MESSAGE;
	HANDLE waitEvents[2] = { exitEvent, hasMessagesEvent };
	char messageLengthBuffer[4];

	while (true)
	{
		auto result = WaitForMultipleObjects(2, waitEvents, false, INFINITE);
		if (result == WAIT_OBJECT_0)
			break;
		else if (result == WAIT_OBJECT_0 + 1)
		{
			std::unique_lock<std::mutex> lock(mutex);
			while (!pendingMessages.empty())
			{
				auto message = StringHelper::ToAString(pendingMessages.front());

				int size = message.size();
				messageLengthBuffer[0] = ((char*)(&size))[0];
				messageLengthBuffer[1] = ((char*)(&size))[1];
				messageLengthBuffer[2] = ((char*)(&size))[2];
				messageLengthBuffer[3] = ((char*)(&size))[3];

				Write(messageLengthBuffer, 4);
				Write(message.data(), message.size());

				pendingMessages.pop();
			}
		}
	}
	if (shutdownCallback)
		shutdownCallback(this);
}
