#include "WebAppRPCProcessor.h"
#include "MessageProcessor/MessageProcessor.h"
#include "MessageProcessor/StdInOutMessageProcessor.h"
#include "../Browser/BrowserClient.h"
#include "../Thread/ManualResetEvent.h"
#include "../Utils/StringHelper.h"
#include "../Utils/Profiler.h"

#include "../Browser/SiteKioskSchemeHandlerFactory.h"

#include "include/cef_scheme.h"
#include "include/cef_app.h"
#include "include/cef_origin_whitelist.h"

#include <iostream>
#include <json\json.h>


const std::wstring WebAppRPCProcessor::AppCommandPrefix = L"GDF9fFH65:";
const std::wstring WebAppRPCProcessor::ObjectModelMessagePrefix = L"C4A8EF1F0:";
const std::wstring WebAppRPCProcessor::ViewMessagePrefix = L"BDF64H7F8:";


WebAppRPCProcessor::WebAppRPCProcessor()
	: RPCProcessor(MessageLoggingLevel::ResponseAndRequests)
	, messageProcessor(new StdInOutMessageProcessor())
	, appCommands(*this)
	, viewCommands(*this)
{
	CefRefPtr<CefSchemeHandlerFactory> factory = new SiteKioskSchemeHandlerFactory();
	CefRegisterSchemeHandlerFactory("sk", "siteKiosk", factory);

	messageProcessor->AddMessageHandler(WebAppRPCProcessor::AppCommandPrefix, std::bind(&WebAppRPCProcessor::OnAppMessage, this, std::placeholders::_1));
	messageProcessor->AddMessageHandler(WebAppRPCProcessor::ViewMessagePrefix, std::bind(&WebAppRPCProcessor::OnViewMessage, this, std::placeholders::_1));

	appCommands.Init();
	viewCommands.Init();

	messageProcessor->Start();
}


WebAppRPCProcessor::~WebAppRPCProcessor()
{
	std::unique_lock<std::mutex> lock(mutex);
	messageProcessor.reset();
}


void WebAppRPCProcessor::RegisterNonBlockingAppCommand(const std::wstring& method, RPCFunctionNonBlocking commandFunction)
{
	std::unique_lock<std::mutex> lock(mutex);

	Handler handler(method, commandFunction);
	registeredAppCommands.push_back(handler);
}


void WebAppRPCProcessor::RegisterAppCommand(const std::wstring& method, RPCFunction commandFunction)
{
	std::unique_lock<std::mutex> lock(mutex);

	Handler handler(method, commandFunction);
	registeredAppCommands.push_back(handler);
}


void WebAppRPCProcessor::RegisterNonBlockingViewCommand(const std::wstring& method, RPCFunctionNonBlocking commandFunction)
{
	std::unique_lock<std::mutex> lock(mutex);

	Handler handler(method, commandFunction);
	registeredViewCommands.push_back(handler);
}


void WebAppRPCProcessor::RegisterViewCommand(const std::wstring& method, RPCFunction commandFunction)
{
	std::unique_lock<std::mutex> lock(mutex);

	Handler handler(method, commandFunction);
	registeredViewCommands.push_back(handler);
}


void WebAppRPCProcessor::Shutdown()
{
	std::unique_lock<std::mutex> lock(mutex);

	messageProcessor->Shutdown();
	RPCProcessor::Shutdown();
}


bool WebAppRPCProcessor::OnAppMessage(const std::wstring& message)
{
	PERFORMANCER;

	JsonHandlerInfo jsonHandlerInfo;
	jsonHandlerInfo.Handlers = registeredAppCommands;
	jsonHandlerInfo.ResponsePrefix = WebAppRPCProcessor::AppCommandPrefix;

	return HandleJsonRequest(jsonHandlerInfo, message);
}


bool WebAppRPCProcessor::OnViewMessage(const std::wstring& message)
{
	PERFORMANCER;

	auto values = StringHelper::Split(message, L':', 2);
	int targetId = values.size() == 2 ? std::stoi(values[0]) : -1;
	auto realMessage = values.size() == 2 ? values[1] : values[0];

	JsonHandlerInfo jsonHandlerInfo;
	jsonHandlerInfo.Handlers = registeredViewCommands;
	jsonHandlerInfo.ResponsePrefix = WebAppRPCProcessor::ViewMessagePrefix + std::to_wstring(targetId) + L":";

	return HandleJsonRequest(jsonHandlerInfo, realMessage, targetId);
}

std::wstring WebAppRPCProcessor::PrettifyMessagePrefix(const std::wstring& prefix)
{
	auto values = StringHelper::Split(prefix, L':', 2);
	auto realPrefix = values[0] + L":";
	if (realPrefix == WebAppRPCProcessor::AppCommandPrefix)
		realPrefix = L"App Command";
	else if (realPrefix == WebAppRPCProcessor::ObjectModelMessagePrefix)
		realPrefix = L"Object Model Call";
	else if (realPrefix == WebAppRPCProcessor::ViewMessagePrefix)
		realPrefix = L"View Message";
	if (values.size() == 1)
		return realPrefix;
	return realPrefix + L" (View " + values[1] + L")";
}
