#include "ThreadPool.h"
#include "../Utils/ThreadHelper.h"
#include <sstream>
#include <algorithm>


ThreadPool::ThreadPool(const std::string& name, unsigned int minThreadCount, unsigned int maxThreadCount, DWORD threadTimeOut)
	: minThreadCount(minThreadCount <= 0 ? 1 : minThreadCount)
	, maxThreadCount(maxThreadCount < minThreadCount ? minThreadCount : maxThreadCount)
	, threadTimeOut(threadTimeOut)
	, name(name)
	, exitEvent(false)
	, tasksAvailableEvent(false)
{
	std::unique_lock<std::mutex> lock(mutex);

	for (unsigned int i = 0; i < minThreadCount; ++i)
	{
		auto threadPtr = std::shared_ptr<Thread>(new Thread(name, *this, true));
		threads.push_back(threadPtr);
	}
}


ThreadPool::~ThreadPool()
{
	std::unique_lock<std::mutex> lock(mutex);

	exitEvent.Set();
	threads.clear();
}


void ThreadPool::Dispatch(const std::function<void()>& method)
{
	std::unique_lock<std::mutex> lock(mutex);

	CleanUp();

	if (threads.size() < maxThreadCount && AllThreadsOccupied())
		SpawnNewThread();

	availableTasks.push(method);
	tasksAvailableEvent.Set();
}


bool ThreadPool::AllThreadsOccupied()
{
	for (const auto& thread : threads)
		if (thread->IsIdle())
			return false;
	return true;
}


void ThreadPool::SpawnNewThread()
{
	auto threadPtr = std::shared_ptr<Thread>(new Thread(name, *this, false));
	threads.push_back(threadPtr);
}


void ThreadPool::CleanUp()
{
	for (unsigned int i = 0; i < threads.size();)
	{
		if (threads[i]->IsDead())
			threads.erase(threads.begin() + i);
		else
			++i;
	}
}


std::function<void()> ThreadPool::ConsumeEvent()
{
	std::unique_lock<std::mutex> lock(mutex);

	if (availableTasks.empty())
		return nullptr;

	auto task = availableTasks.front();
	availableTasks.pop();

	if (availableTasks.empty())
		tasksAvailableEvent.Reset();

	return task;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
ThreadPool::Thread::Thread(const std::string& name, ThreadPool& owner, bool isCoreThread)
	: owner(owner)
	, isDead(false)
	, isCoreThread(isCoreThread)
	, isWorkingEvent(false)
{
	thread = std::unique_ptr<std::thread>(new std::thread(&ThreadPool::Thread::ProcessMessagesAsync, this));
	ThreadHelper::SetThreadName(*thread, "ThreadPool '" + name + "' Thread");
}


ThreadPool::Thread::~Thread()
{
	thread->join();
}


bool ThreadPool::Thread::IsIdle()
{
	return isWorkingEvent.Wait(0) != WAIT_OBJECT_0;
}


void ThreadPool::Thread::ProcessMessagesAsync()
{
	HANDLE waitEvents[2] = { owner.exitEvent, owner.tasksAvailableEvent };

	while (true)
	{
		auto result = WaitForMultipleObjects(2, waitEvents, false, owner.threadTimeOut);
		if (result == WAIT_TIMEOUT)
		{
			isDead = !isCoreThread;
			if (isDead)
				break;
			else
				continue;
		}
		else if (result == WAIT_OBJECT_0)
			break;
		else if (result == WAIT_OBJECT_0 + 1)
		{
			auto method = owner.ConsumeEvent();
			if (method == nullptr)
				continue;

			isWorkingEvent.Set();
			method();
			isWorkingEvent.Reset();
		}
		else
		{
			isDead = true;

			std::wstringstream ss;
			ss << "ThreadPool Thread will end, but this was not expected. ErrorCode: " << GetLastError() << std::endl;
			OutputDebugString(ss.str().c_str());

			break;
		}
	}
}
