#include "DisplayComponent.h"
#include "LayoutComponent.h"
#include "../View.h"
#include "../../Browser/BrowserClient.h"
#include "../../Utils/StringHelper.h"
#include "../../Thread/ManualResetEvent.h"

#include "include/base/cef_bind.h"
#include "include/cef_app.h"
#include "include/wrapper/cef_closure_task.h"

DisplayComponent::DisplayComponent(View& view)
	: Component(view)
	, hasBorders(true)
	, browserHandle(nullptr)
	, zoom(1.0f)
{
}


void DisplayComponent::UpdateStyles()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	baseStyle = GetWindowLong(view.GetHandle(), GWL_STYLE);
	baseExStyle = GetWindowLong(view.GetHandle(), GWL_EXSTYLE);

	SetHasBorder(hasBorders);
}


void DisplayComponent::SetTitle(const std::wstring& title)
{
	view.WaitForReady();

	SetWindowText(view.GetHandle(), title.c_str());
}


std::wstring DisplayComponent::GetTitle()
{
	view.WaitForReady();

	auto length = GetWindowTextLengthW(view.GetHandle());
	if (length == 0)
		return std::wstring(L"");

	wchar_t* buf = new wchar_t[length + 1];
	wmemset(buf, L'\0', length + 1);
	GetWindowTextW(view.GetHandle(), buf, length + 1);
	
	std::wstring result = buf;
	
	delete[] buf;

	return result;
}


float DisplayComponent::ConvertToZoomLevel(float zoom)
{
	return log(zoom) / log(1.2f);
}


void DisplayComponent::SetZoom(float zoom)
{
	this->zoom = zoom;
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GetHost()->SetZoomLevel(ConvertToZoomLevel(zoom));
}


float DisplayComponent::GetZoom()
{
	return zoom;
}


void DisplayComponent::SetShowInTaskbar(bool value)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	if (!value)
		baseExStyle |= WS_EX_NOACTIVATE;
	else
		baseExStyle &= ~WS_EX_NOACTIVATE;

	SetWindowLong(view.GetHandle(), GWL_EXSTYLE, baseExStyle);
}


bool DisplayComponent::GetShowInTaskbar()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	return (baseExStyle & WS_EX_NOACTIVATE) != WS_EX_NOACTIVATE;
}


void DisplayComponent::SetIsVisible(bool value)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();


	if (value)
		baseStyle |= WS_VISIBLE;
	else
		baseStyle &= ~WS_VISIBLE;

	ShowWindow(view.GetHandle(), !value ? SW_HIDE : SW_SHOW);
	SetWindowLong(view.GetHandle(), GWL_STYLE, baseStyle);
}


bool DisplayComponent::GetIsVisible()
{
	view.WaitForReady();

	return IsWindowVisible(view.GetHandle()) == TRUE;
}


void DisplayComponent::SetHasBorder(bool hasBorder)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	

	hasBorders = hasBorder;

	if (view.GetHandle() == nullptr)
		return;
	view.WaitForReady();

	if (!hasBorder)
	{
		baseStyle &= ~(WS_CAPTION | WS_THICKFRAME | WS_SYSMENU | WS_MINIMIZEBOX | WS_MAXIMIZEBOX);
		baseExStyle &= ~(WS_EX_DLGMODALFRAME | WS_EX_CLIENTEDGE | WS_EX_STATICEDGE);
	}
	else
	{
		baseStyle |= (WS_CAPTION | WS_THICKFRAME | WS_SYSMENU | WS_MINIMIZEBOX | WS_MAXIMIZEBOX);
		baseExStyle |= (WS_EX_DLGMODALFRAME | WS_EX_CLIENTEDGE | WS_EX_STATICEDGE);
	}

	SetWindowLong(view.GetHandle(), GWL_STYLE, baseStyle);
	SetWindowLong(view.GetHandle(), GWL_EXSTYLE, baseExStyle);
	SetWindowPos(view.GetHandle(), NULL, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


bool DisplayComponent::GetHasBorder()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	return hasBorders;
}


void DisplayComponent::SetBrowserHandle(HWND handle)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	browserHandle = handle;

	view.Layout.SetParent(view.Layout.GetParent());
}


HWND DisplayComponent::GetBrowserHandle()
{
	return browserHandle;
}


void DisplayComponent::ChangeChildStyle(bool isChild)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();


	if (isChild)
	{
		baseStyle |= WS_CHILD;
		baseStyle &= ~WS_OVERLAPPED;
	}
	else
	{
		baseStyle |= WS_OVERLAPPED;
		baseStyle &= ~WS_CHILD;
	}

	SetWindowLong(view.GetHandle(), GWL_STYLE, baseStyle);
}
