#include "ErrorComponent.h"
#include "../../Utils/CommandLine.h"
#include "../../Utils/StringHelper.h"
#include "../View.h"
#include "../../Browser/BrowserClient.h"

#include <iostream>
#include <fstream>
#include <regex>

ErrorComponent::ErrorComponent(View& view)
	: Component(view)
{
	fileBaseDir = CommandLine::GetInstance().GetStringArgument(L"assetsPath");
}


void ErrorComponent::DisplayError(const std::wstring& failedUrl, int errorCode, int frameId)
{
	auto isSpecialError = errorCode == 0;
	if (isSpecialError)
		errorCode = ERR_NAME_NOT_RESOLVED;

	errorCode = RemapErrorCode(errorCode);
	std::wstring errorFileBaseDir = fileBaseDir + L"\\errorPages\\";
	auto errorPage = errorFileBaseDir + std::to_wstring(errorCode);
	auto fallbackPage = errorFileBaseDir + L"error.html";
	std::ifstream htmlTemplateFile(errorPage);
	if (!htmlTemplateFile.is_open())
		htmlTemplateFile = std::ifstream(fallbackPage);

	std::wstring htmlContent;
	if (htmlTemplateFile.is_open())
	{
		std::stringstream sstr;
		sstr << htmlTemplateFile.rdbuf();
		htmlContent = StringHelper::ToWString(sstr.str());
	}
	else
	{
		htmlContent = L"<html><body bgcolor=\"white\"><h2>"
			L"Failed to load URL $URL with error $ErrorText."
			L"</h2></body></html>";
	}

	auto translatedErrorText = view.GetString(std::wstring(L"LoadError.Code(") + std::to_wstring(errorCode) + L")");
	if (translatedErrorText == L"")
		translatedErrorText = std::to_wstring(errorCode);

	std::set<std::wstring> stringsToTranslate;
	do
	{
		htmlContent = StringHelper::Replace(htmlContent, std::wstring(L"$BasePath"), std::wstring(errorFileBaseDir));
		htmlContent = StringHelper::Replace(htmlContent, std::wstring(L"$URL"), std::wstring(failedUrl));
		htmlContent = StringHelper::Replace(htmlContent, std::wstring(L"$ErrorText"), std::wstring(translatedErrorText));
		htmlContent = StringHelper::Replace(htmlContent, std::wstring(L"$ErrorCode"), std::to_wstring(errorCode));

		stringsToTranslate = FindRequestedStrings(htmlContent);
		for (auto stringToTranslate : stringsToTranslate)
		{
			auto translated = view.GetString(stringToTranslate);
			if (translated == L"")
				translated = stringToTranslate;

			htmlContent = StringHelper::Replace(htmlContent, L"$String(" + stringToTranslate + L")", translated);
		}
	} while (!stringsToTranslate.empty());

	CefRefPtr<CefFrame> frame;
	if (frameId == -1)
		frame = view.GetBrowser()->GetMainFrame();
	else
		frame = view.GetBrowser()->GetFrame(frameId);

	frame->LoadString(htmlContent, isSpecialError ? L"http://" + failedUrl : failedUrl);
}


std::set<std::wstring> ErrorComponent::FindRequestedStrings(const std::wstring& input)
{
	std::wregex stringRx(L"\\$String\\(([^)]*)\\)");

	std::set<std::wstring> result;

	auto s = input;
	std::wsmatch m;
	while (std::regex_search(s, m, stringRx))
	{
		if (m.size() == 2)
			result.insert(m[1]);
		s = m.suffix().str();
	}

	return result;
}


int ErrorComponent::RemapErrorCode(int errorCode)
{
	const int allowedErrorCodes[] = { 
		ERR_NONE,
		ERR_FAILED,
		ERR_ABORTED,
		ERR_INVALID_ARGUMENT,
		ERR_INVALID_HANDLE,
		ERR_FILE_NOT_FOUND,
		ERR_TIMED_OUT,
		ERR_FILE_TOO_BIG,
		ERR_UNEXPECTED,
		ERR_ACCESS_DENIED,
		ERR_NOT_IMPLEMENTED,
		ERR_CONNECTION_CLOSED,
		ERR_CONNECTION_RESET,
		ERR_CONNECTION_REFUSED,
		ERR_CONNECTION_ABORTED,
		ERR_CONNECTION_FAILED,
		ERR_NAME_NOT_RESOLVED,
		ERR_INTERNET_DISCONNECTED,
		ERR_SSL_PROTOCOL_ERROR,
		ERR_ADDRESS_INVALID,
		ERR_ADDRESS_UNREACHABLE,
		ERR_SSL_CLIENT_AUTH_CERT_NEEDED,
		ERR_TUNNEL_CONNECTION_FAILED,
		ERR_NO_SSL_VERSIONS_ENABLED,
		ERR_SSL_VERSION_OR_CIPHER_MISMATCH,
		ERR_SSL_RENEGOTIATION_REQUESTED,
		ERR_CERT_COMMON_NAME_INVALID,
		ERR_CERT_DATE_INVALID,
		ERR_CERT_AUTHORITY_INVALID,
		ERR_CERT_CONTAINS_ERRORS,
		ERR_CERT_NO_REVOCATION_MECHANISM,
		ERR_CERT_UNABLE_TO_CHECK_REVOCATION,
		ERR_CERT_REVOKED,
		ERR_CERT_INVALID,
		ERR_CERT_END,
		ERR_INVALID_URL,
		ERR_DISALLOWED_URL_SCHEME,
		ERR_UNKNOWN_URL_SCHEME,
		ERR_TOO_MANY_REDIRECTS,
		ERR_UNSAFE_REDIRECT,
		ERR_UNSAFE_PORT,
		ERR_INVALID_RESPONSE,
		ERR_INVALID_CHUNKED_ENCODING,
		ERR_METHOD_NOT_SUPPORTED,
		ERR_UNEXPECTED_PROXY_AUTH,
		ERR_EMPTY_RESPONSE,
		ERR_RESPONSE_HEADERS_TOO_BIG ,
		ERR_CACHE_MISS,
		ERR_INSECURE_RESPONSE,
		-118 // connection timed out
	};
	bool exists = std::find(std::begin(allowedErrorCodes), std::end(allowedErrorCodes), errorCode) != std::end(allowedErrorCodes);
	return exists ? errorCode : ERR_FAILED;
}
