#include "EventComponent.h"
#include "../../Browser/BrowserClient.h"
#include "../../Messages/BrowserRPCProcessor.h"
#include "../../Utils/StringHelper.h"
#include "../../Utils/Profiler.h"
#include "../View.h"
#include "../Components/DisplayComponent.h"
#include "../Components/RequestComponent.h"
#include "../Components/LayoutComponent.h"
#include "../Components/PingComponent.h"
#include "../../resource.h"

#include <windows.h>


EventComponent::EventComponent(View& view)
	: Component(view)
	, editableFrame(-1)
{
}


void EventComponent::OnBrowserCreated(CefRefPtr<CefBrowser> browser)
{
	PERFORMANCER;

	view.SetBrowser(browser);

	view.Display.SetBrowserHandle(browser->GetHost()->GetWindowHandle());

	auto instance = (HINSTANCE)GetModuleHandle(nullptr);
	auto icon = (LONG)LoadIcon(instance, MAKEINTRESOURCE(IDI_ICON));

	SetClassLong(view.GetHandle(), GCL_HICON, icon);

	auto parentId = view.Layout.GetParent();
	view.Layout.SetParent(parentId);

	if (view.IsPopup())
	{
		auto creatorView = BrowserClient::GetInstance().GetViewManager().GetView(view.GetPopupCreatorId());
		if (creatorView != nullptr)
		{
			creatorView->Events.OnPopupViewCreated(view.GetId(), view.GetHandle());
		}
	}

	browser->GetHost()->NotifyMoveOrResizeStarted();
	browser->GetHost()->WasResized();

#if DEBUG
	std::wstringstream outputStringStream;

	outputStringStream << CEF_APP_LOG_PREFIX;
	outputStringStream << "Assigned View: " << view.GetId() << " to Browser: " << browser->GetIdentifier() << std::endl;

	OutputDebugString(outputStringStream.str().c_str());
#endif
}


void EventComponent::OnLoadStart(CefRefPtr<CefFrame> frame)
{
	PERFORMANCER;

	view.Display.SetZoom(view.Display.GetZoom());
	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnLoadStart(view.GetId(), frame);
}


void EventComponent::OnLoadEnd(CefRefPtr<CefFrame> frame, int httpStatusCode)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnLoadEnd(view.GetId(), frame, httpStatusCode);
}


void EventComponent::OnLoadError(CefRefPtr<CefFrame> frame, int errorCode, const std::wstring& errorText, const std::wstring& failedUrl)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnLoadError(view.GetId(), frame, errorCode, errorText, failedUrl);
}


void EventComponent::OnLoadingStateChange(bool isLoading, bool canGoBack, bool canGoForward)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnLoadingStateChange(view.GetId(), isLoading, canGoBack, canGoForward);
}


bool EventComponent::OnCheckMustCancelRequest(CefRefPtr<CefFrame> frame, const std::wstring& url, const std::wstring& content, CheckMustCancelRequestFlags checkFlags)
{
	PERFORMANCER;

	auto browserRpc = view.GetBrowserRPCProcessor();
	if (browserRpc == nullptr)
		return true;
	return BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnCheckMustCancelRequest(view.GetId(), frame, url, content, checkFlags, browserRpc->GetExitEventHandle());
}


bool EventComponent::OnFileUploadRequest(CefRefPtr<CefFrame> frame, const std::wstring& url, const std::vector<std::wstring>& content)
{
	PERFORMANCER;

	auto browserRpc = view.GetBrowserRPCProcessor();
	if (browserRpc == nullptr)
		return true;
	return BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnFileUploadRequest(view.GetId(), frame, url, content, browserRpc->GetExitEventHandle());
}


bool EventComponent::OnBeforeNavigation(const std::wstring& url, bool isMainFrame, int frameId)
{
	PERFORMANCER;

	auto browserRpc = view.GetBrowserRPCProcessor();
	if (browserRpc == nullptr)
		return true;
	return BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnBeforeNavigation(view.GetId(), url, isMainFrame, frameId, browserRpc->GetExitEventHandle());
}


bool EventComponent::OnBeforeBrowse(const std::wstring& url, bool isMainFrame, bool isRedirect, int frameId)
{
	PERFORMANCER;

	return BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnBeforeBrowse(view.GetId(), url, isMainFrame, isRedirect, frameId);
}


void EventComponent::OnContextCreated(int64 frameId)
{
	PERFORMANCER;

	if (view.GetBrowser()->GetMainFrame()->GetIdentifier() == frameId)
		BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnContextCreated(view.GetId());
}


void EventComponent::OnContextReleased(int64 frameId)
{
	PERFORMANCER;

	if (view.GetBrowser()->GetMainFrame()->GetIdentifier() == frameId)
	{
		auto browserRpc = view.GetBrowserRPCProcessor();
		if (browserRpc == nullptr)
			return;
		BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnContextReleased(view.GetId(), browserRpc->GetExitEventHandle());
	}
}


void EventComponent::OnTitleChanged(const std::wstring& title)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnTitleChanged(view.GetId(), title);
}


void EventComponent::OnGotFocus(bool isForegroundWindow)
{
	PERFORMANCER;

	//if (isForegroundWindow)
		BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnGotFocus(view.GetId());
}


void EventComponent::OnEditableNodeFocusChangedHandler(bool isEditable, int64 frameId)
{
	PERFORMANCER;

	if (isEditable)
		editableFrame = frameId;
	else
		editableFrame = -1;

	BrowserClient::GetInstance().GetRPCProcessor().GetAppCommands().OnEditableNodeFocusChanged(isEditable);
}


void EventComponent::OnFaviconURLChange(const std::vector<std::wstring>& iconUrls)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnFaviconURLChange(view.GetId(), iconUrls);
}


void EventComponent::OnPopupViewCreated(int popupViewId, HANDLE popupWindowHandle)
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnPopupViewCreated(view.GetId(), popupViewId, (int)popupWindowHandle);
}


void EventComponent::OnViewIsReady()
{
	PERFORMANCER;

	view.GetViewReadyEvent().Set();
	view.Pinger.Pong();
}


void EventComponent::OnRendererCreated()
{
	PERFORMANCER;

	view.Requests.CancelRequests();
}


void EventComponent::OnActivated()
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnActivated(view.GetId());
}


void EventComponent::OnDeactivated()
{
	PERFORMANCER;

	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnDeactivated(view.GetId());
}
