#include "InteractionComponent.h"
#include "DisplayComponent.h"
#include "LayoutComponent.h"
#include "EventComponent.h"
#include "../View.h"
#include "../ViewManager.h"
#include "../../Browser/BrowserClient.h"
#include "../../Messages/BrowserRPCProcessor.h"
#include "../../Utils/CommandLine.h"


InteractionComponent::InteractionComponent(View& view)
	: Component(view)
	, isClosable(true)
	, allowContextMenu(false)
	, allowPrinting(false)
	, popupHandlingMode(PopupHandlingMode::DontOpen)
{
	if (CommandLine::GetInstance().GetBoolArgument(L"handlePopupWindowsAsLinks"))
		popupHandlingMode = PopupHandlingMode::OpenInCurrentView;
	else if (CommandLine::GetInstance().GetBoolArgument(L"allowPopupWindows"))
		popupHandlingMode = PopupHandlingMode::OpenAsNewPopup;
}


void SetActiveWindowInternal(HWND hWnd)
{
	DWORD lockTimeOut = 0;
	DWORD dwThisTID = ::GetCurrentThreadId(),
		dwCurrTID = ::GetWindowThreadProcessId(hWnd, 0);

	::AttachThreadInput(dwThisTID, dwCurrTID, TRUE);
	::SetActiveWindow(hWnd);
	::AttachThreadInput(dwThisTID, dwCurrTID, FALSE);
}


void InteractionComponent::Focus()
{
	view.WaitForReady();

	auto rootHandle = BrowserClient::GetInstance().GetViewManager().GetRootView(view.GetId())->GetHandle();
	SetActiveWindowInternal(rootHandle);

	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GetHost()->SetFocus(true);

	BrowserClient::GetInstance().GetViewManager().SetLastFocusedView(view.GetId());
}


void InteractionComponent::Activate()
{
	view.WaitForReady();

	SetActiveWindowInternal(view.GetHandle());
}


void InteractionComponent::SetIsCloseable(bool value)
{
	isClosable = value;
}


bool InteractionComponent::GetIsCloseable()
{
	return isClosable;
}


PopupHandlingMode InteractionComponent::GetPopupHandling() const
{
	return popupHandlingMode;
}


void InteractionComponent::SetPopupHandling(PopupHandlingMode mode)
{
	popupHandlingMode = mode;
}


void InteractionComponent::SetAllowInteraction(bool value)
{
	view.WaitForReady();

	EnableWindow(view.GetHandle(), value);
}


bool InteractionComponent::GetAllowInteraction()
{
	view.WaitForReady();

	return IsWindowEnabled(view.GetHandle()) == TRUE;
}


void InteractionComponent::LoadUrl(const std::wstring& url, int frameId)
{
	CefRefPtr<CefFrame> frame;

	auto browser = view.GetBrowser();
	if (browser != nullptr)
	{
		if (frameId == -1)
			frame = browser->GetMainFrame();
		else
			frame = browser->GetFrame(frameId);
	}

	if (frame == nullptr)
		return;

	if (view.Events.OnCheckMustCancelRequest(frame, url, L"", CheckMustCancelRequestFlags::CheckScheme))
		return;

	frame->LoadURL(url);
}


std::wstring InteractionComponent::GetUrl()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		return browser->GetMainFrame()->GetURL();
	return L"about:blank";
}

void InteractionComponent::Reload()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->Reload();
}


bool InteractionComponent::GetCanGoForward()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		return browser->CanGoForward();
	return false;
}


void InteractionComponent::Forward()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GoForward();
}


bool InteractionComponent::GetCanGoBack()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		return browser->CanGoBack();
	return false;
}

void InteractionComponent::Back()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GoBack();
}


void InteractionComponent::Stop()
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->StopLoad();
}


void InteractionComponent::SetAllowContextMenu(bool value)
{
	allowContextMenu = value;
}


bool InteractionComponent::GetAllowContextMenu()
{
	return allowContextMenu;
}


void InteractionComponent::SetAllowPrinting(bool value)
{
	allowPrinting = value;
}


bool InteractionComponent::GetAllowPrinting()
{
	return allowPrinting;
}


void InteractionComponent::Print(const std::wstring& jobName)
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GetHost()->Print();
}


void InteractionComponent::ExecuteScript(const std::wstring& script)
{
	auto browser = view.GetBrowser();
	if (browser != nullptr)
		browser->GetMainFrame()->ExecuteJavaScript(script, "file:///", 0);
}


void InteractionComponent::StartFullscreen()
{
	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnFullscreenChanged(view.GetId(), true);
}


void InteractionComponent::ExitFullscreen()
{
	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnFullscreenChanged(view.GetId(), false);
}


void InteractionComponent::CheckNonClientAreaClick(int x, int y, HWND originalHwnd)
{
	for (const auto& pair : nonClientAreas)
	{
		auto type = pair.first;
		auto rect = pair.second;
		if (x >= rect.left && x <= rect.right &&
			y >= rect.top && y <= rect.bottom)
		{
			switch (type)
			{
			case NonClientAreaType::Caption:
					ReleaseCapture();
					SendMessage(view.GetHandle(), WM_NCLBUTTONDOWN, HTCAPTION, 0);
					return;
			}
		}
	}
}


void InteractionComponent::SetNonClientArea(NonClientAreaType type, int x, int y, int width, int height)
{
	RECT rect;
	rect.left = x;
	rect.top = y;
	rect.right = x + width;
	rect.bottom = y + height;
	nonClientAreas[type] = rect;
}
