#include "LayoutComponent.h"
#include "DisplayComponent.h"
#include "../View.h"
#include "../../Browser/BrowserClient.h"
#include <climits>


LayoutComponent::LayoutComponent(View& view)
	: Component(view)
	, parentViewId(-1)
	, zIndex(0)
	, minWidth(0)
	, minHeight(0)
	, maxWidth(INT_MAX)
	, maxHeight(INT_MAX)
{
}


void LayoutComponent::SetPosition(int x, int y, bool isRelativePosition)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	if (!isRelativePosition)
	{
		auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
		if (parent != nullptr)
		{
			RECT rect;
			GetWindowRect(view.GetHandle(), &rect);

			rect.left = x;
			rect.top = y;

			ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.left);

			x = rect.left;
			y = rect.top;
		}
	}

	SetWindowPos(view.GetHandle(), NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


void LayoutComponent::SetSize(int width, int height)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	SetWindowPos(view.GetHandle(), NULL, 0, 0, width, height, SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


void LayoutComponent::SetPositionAndSize(int x, int y, int width, int height, bool isRelativePosition)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	if (!isRelativePosition)
	{
		auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
		if (parent != nullptr)
		{
			RECT rect;
			GetWindowRect(view.GetHandle(), &rect);

			rect.left = x;
			rect.top = y;

			ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.left);

			x = rect.left;
			y = rect.top;
		}
	}

	SetWindowPos(view.GetHandle(), NULL, x, y, width, height, SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


void LayoutComponent::SetWidth(int width)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, width, rect.bottom - rect.top, SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetWidth()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	return rect.right - rect.left;
}


void LayoutComponent::SetHeight(int height)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, rect.right - rect.left, height, SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetHeight()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	return rect.bottom - rect.top;
}


void LayoutComponent::SetLeft(int left)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent != nullptr)
	{
		rect.left = left;
		ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.left);
		ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.right);

		SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
	}
	else
		SetWindowPos(view.GetHandle(), NULL, left, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetLeft()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	return rect.left;
}


void LayoutComponent::SetRelativeLeft(int left)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent == nullptr)
		return SetLeft(left);

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	ScreenToClient(parent->GetHandle(), (LPPOINT)&rect);

	rect.left = left;

	SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetRelativeLeft()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent == nullptr)
		return GetLeft();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	MapWindowPoints(view.GetHandle(), parent->GetHandle(), (LPPOINT)&rect, 1);

	return rect.left;
}


void LayoutComponent::SetTop(int top)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent != nullptr)
	{
		rect.top = top;
		ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.left);
		ScreenToClient(parent->GetHandle(), (LPPOINT)&rect.right);

		SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
	}
	else
		SetWindowPos(view.GetHandle(), NULL, rect.left, top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetTop()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	return rect.top;
}


void LayoutComponent::SetRelativeTop(int top)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent == nullptr)
		return SetTop(top);

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	ScreenToClient(parent->GetHandle(), (LPPOINT)&rect);

	rect.top = top;

	SetWindowPos(view.GetHandle(), NULL, rect.left, rect.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOOWNERZORDER | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE);
}


int LayoutComponent::GetRelativeTop()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	view.WaitForReady();

	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	if (parent == nullptr)
		return GetTop();

	RECT rect;
	GetWindowRect(view.GetHandle(), &rect);
	MapWindowPoints(view.GetHandle(), parent->GetHandle(), (LPPOINT)&rect, 1);

	return rect.top;
}


void LayoutComponent::SetParent(int newParentViewId)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	if (newParentViewId == view.GetId())
		newParentViewId = -1;
	parentViewId = newParentViewId;

	if (view.GetHandle() == nullptr)
		return;
	view.WaitForReady();
	
	auto parent = BrowserClient::GetInstance().GetViewManager().GetView(parentViewId);
	view.Display.ChangeChildStyle(parent != nullptr);

	if (parent != nullptr)
	{
		HWND parentHandle = parent->GetHandle();
		::SetParent(view.GetHandle(), parent->GetHandle());
	}
	else
	{
		::SetParent(view.GetHandle(), nullptr);

		auto rootView = BrowserClient::GetInstance().GetViewManager().GetRootView();
		if (rootView != nullptr && rootView.get() != &view)
			::SetWindowLong(view.GetHandle(), GWL_HWNDPARENT, (long)rootView->GetHandle());
	}

	UpdateWindow(view.GetHandle());
}


int LayoutComponent::GetParent()
{
	//std::unique_lock<std::recursive_mutex> lock(mutex);
	return parentViewId;
}


void LayoutComponent::SetZIndex(int zIndex)
{
	this->zIndex = zIndex;
	BrowserClient::GetInstance().GetViewManager().OnZOrderChanged(view.GetId());
}


int LayoutComponent::GetZIndex()
{
	return zIndex;
}


void LayoutComponent::SetMinSize(int width, int height)
{
	minWidth = width;
	minHeight = height;
}


void LayoutComponent::SetMaxSize(int width, int height)
{
	maxWidth = width;
	maxHeight = height;
}


SIZE LayoutComponent::GetMinSize() const
{
	SIZE result;
	result.cx = minWidth;
	result.cy = minHeight;
	return result;
}


SIZE LayoutComponent::GetMaxSize() const
{
	SIZE result;
	result.cx = maxWidth;
	result.cy = maxHeight;
	return result;
}
