#include "RequestComponent.h"
#include "../View.h"
#include "../../Browser/BrowserClient.h"
#include "../../Messages/Cancelable.h"
#include "../../Utils/Profiler.h"


RequestComponent::RequestComponent(View& view)
	: Component(view)
	, contentFilterLevel(0)
{
}

RequestComponent::~RequestComponent()
{
	CancelRequests();
}


void RequestComponent::SetContentFilterLevel(int level)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	contentFilterLevel = level;
}


int RequestComponent::GetContentFilterLevel() const
{
	return contentFilterLevel;
}


void RequestComponent::RequestAuthCredentials(bool isProxy, const std::wstring& host, int port, const std::wstring& realm, const std::wstring& scheme, const std::function<void(const AuthenticationResult&)>& callback)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	if (authRequestCancelable != nullptr)
		authRequestCancelable->Cancel();
	authRequestCancelable = BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().RequestAuthCredentials(view.GetId(), isProxy, host, port, realm, scheme, callback);
}


void RequestComponent::RequestQuota(const std::wstring& origin, int64 newSize, const std::function<void(bool)>& callback)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	if (quotaRequestCancelable != nullptr)
		quotaRequestCancelable->Cancel();

	quotaRequestCancelable = BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().RequestQuota(view.GetId(), view.GetBrowser()->GetMainFrame()->GetURL(), newSize, callback);
}


void RequestComponent::RequestDownload(const std::wstring& url, int64 totalBytes, const std::function<void(const DownloadRequestResult&)>& callback)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	if (downloadRequestCancelable != nullptr)
		downloadRequestCancelable->Cancel();
	downloadRequestCancelable = BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().RequestDownload(view.GetId(), url, totalBytes, callback);
}


void RequestComponent::DisplayDialog(const std::wstring& origin, const std::wstring& accept_lang, int dialogType, const std::wstring& message, const std::wstring& defaultPromptText, const std::function<void(const DisplayDialogResult&)>& callback)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	if (displayDialogCancelable != nullptr)
		displayDialogCancelable->Cancel();
	displayDialogCancelable = BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().DisplayDialog(view.GetId(), view.GetBrowser()->GetMainFrame()->GetURL(), accept_lang, dialogType, message, defaultPromptText, callback);
}


void RequestComponent::DisplayFileDialog(int mode, const std::wstring& title, const std::wstring& defaultFilePath, const std::vector<std::wstring>& acceptFilters, int selectedAcceptFilter, const std::function<void(const DisplayFileDialogResult&)>& callback)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	if (displayFileDialogCancelable != nullptr)
		displayFileDialogCancelable->Cancel();
	displayFileDialogCancelable = BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().DisplayFileDialog(view.GetId(), mode, title, defaultFilePath, acceptFilters, selectedAcceptFilter, callback);
}


void RequestComponent::CancelRequests()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	if (quotaRequestCancelable != nullptr)
		quotaRequestCancelable->Cancel();
	quotaRequestCancelable = nullptr;

	if (authRequestCancelable != nullptr)
		authRequestCancelable->Cancel();
	authRequestCancelable = nullptr;

	if (displayDialogCancelable != nullptr)
		displayDialogCancelable->Cancel();
	displayDialogCancelable = nullptr;

	if (displayFileDialogCancelable != nullptr)
		displayFileDialogCancelable->Cancel();
	displayFileDialogCancelable = nullptr;
}
