#include "View.h"
#include "ViewManager.h"

#include "../Browser/SiteKioskSchemeHandlerFactory.h"
#include "../Browser/BrowserClient.h"
#include "../Browser/Handler/RequestContextHandler.h"
#include "../Messages/BrowserRPCProcessor.h"
#include "../Utils/ThreadHelper.h"
#include "../Utils/Profiler.h"

#include "Components/LayoutComponent.h"
#include "Components/DisplayComponent.h"
#include "Components/InteractionComponent.h"
#include "Components/RequestComponent.h"
#include "Components/EventComponent.h"
#include "Components/PingComponent.h"
#include "Components/ErrorComponent.h"

#include <json/json.h>
#include <sstream>
#include <string>
#include <algorithm>
#include "../resource.h"

#include "include/wrapper/cef_helpers.h"
#include "include/cef_task.h"
#include "include/base/cef_bind.h"
#include "include\wrapper\cef_closure_task.h"

// We use the global request context because of the cookie manager, which doesnt use our websocket settings.
#define USE_GLOBAL_REQUESTCONTEXT


View::View(int id)
	: id(id)
	, allowClose(false)
	, isMainView(id == 0)
	, popupCreatorId(-1)
	, layoutComponent(new LayoutComponent(*this))
	, displayComponent(new DisplayComponent(*this))
	, interactionComponent(new InteractionComponent(*this))
	, requestComponent(new RequestComponent(*this))
	, eventComponent(new EventComponent(*this))
	, pingComponent(new PingComponent(*this))
	, errorComponent(new ErrorComponent(*this))
	, Layout(*layoutComponent)
	, Display(*displayComponent)
	, Interaction(*interactionComponent)
	, Requests(*requestComponent)
	, Events(*eventComponent)
	, Pinger(*pingComponent)
	, Error(*errorComponent)
	, viewReady(false)
	, browserReady(false)
	, windowHandle(nullptr)
	, isClosing(false)
{
#if DEBUG
	std::wstringstream outputStringStream;

	outputStringStream << CEF_APP_LOG_PREFIX;
	outputStringStream << "Created View: " << id << std::endl;
	OutputDebugString(outputStringStream.str().c_str());
#endif
}


View::~View()
{
	if (windowHandle != nullptr)
		PostMessage(windowHandle, WM_CLOSE, 0, 0);

	if (windowThread != nullptr)
	{
		windowThread->detach();
		windowThread = nullptr;
	}

	layoutComponent = nullptr;
	displayComponent = nullptr;
	interactionComponent = nullptr;
	requestComponent = nullptr;
	eventComponent = nullptr;
	pingComponent = nullptr;
	browser = nullptr;
	
	BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().OnClosed(id);
	BrowserClient::GetInstance().GetViewManager().OnViewDestructed(id);
}


void View::Init(const CreateViewParams &params)
{
	PERFORMANCER;

	createParams = params;
	CreateContainerWindow(false);
}


void View::Init(int creatorId, CefWindowInfo& windowInfo, CefRefPtr<CefBrowser> browser, CefBrowserSettings& settings)
{
	PERFORMANCER;

	popupCreatorId = creatorId;
	CreateContainerWindow(true);

	windowInfo.SetAsPopup(GetHandle(), L"");
	RECT windowRect;
	GetClientRect(GetHandle(), &windowRect);
	windowInfo.SetAsChild(GetHandle(), windowRect);
}


void View::SetBrowser(CefRefPtr<CefBrowser> browser)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	this->browser = browser;
	browserReady.Set();
}


CefRefPtr<CefBrowser> View::GetBrowser()
{
	return browser;
}


int View::GetId()
{
	return id;
}


std::shared_ptr<BrowserRPCProcessor> View::GetBrowserRPCProcessor()
{
	return BrowserRPCProcessor::Get(browser);
}


ManualResetEvent& View::GetViewReadyEvent()
{
	return viewReady;
}


std::wstring View::GetString(const std::wstring& stringId)
{
	return BrowserClient::GetInstance().GetRPCProcessor().GetViewCommands().GetString(id, stringId);
}


void View::Close()
{
	Close(true);
}


void View::Close(bool isUserInstigated)
{
	PERFORMANCER;

	std::unique_lock<std::recursive_mutex> lock(mutex);
	Interaction.SetIsCloseable(true);

	Layout.SetParent(-1); // We need this, Cef will send close commands downwards so we stop this ridiculous behaviour by deparenting here.

	if (windowHandle != nullptr)
		PostMessage(windowHandle, WM_CLOSE, 0, 0);
	else if (!isClosing)
	{
		isClosing = true;
		OutputDebugString(L"Browserless view will be removed.");
		auto& viewManager = BrowserClient::GetInstance().GetViewManager();
		viewManager.RemoveView(viewManager.GetView(id));
	}
}


void View::CreateContainerWindow(bool isPopupWindow)
{
	PERFORMANCER;

	{
		std::unique_lock<std::recursive_mutex> lock(mutex);

		Requests.SetContentFilterLevel(createParams.ContentFilterLevel);
		Display.SetHasBorder(createParams.ShowBorders);
		Interaction.SetIsCloseable(createParams.IsClosable);
		Layout.SetParent(createParams.ParentId);
		Layout.SetMinSize(createParams.MinWidth, createParams.MinHeight);
		Layout.SetMaxSize(createParams.MaxWidth, createParams.MaxHeight);
		Layout.SetZIndex(createParams.ZIndex);

		windowThread = std::unique_ptr<std::thread>(new std::thread(&View::ThreadProc, this));
		ThreadHelper::SetThreadName(*windowThread, "View " + std::to_string(id));
	}
	
	BrowserClient::GetInstance().GetViewManager().GetViewExistingEvent().Set();

	ManualResetEvent& exitEvent = BrowserClient::GetInstance().GetShouldQuitEvent();

	if (!isPopupWindow)
	{
		HANDLE waitEvents[2] = { exitEvent, browserReady };
		WaitForMultipleObjects(2, waitEvents, false, INFINITE);
	}
	else
	{
		HANDLE waitEvents[2] = { exitEvent, viewReady };
		WaitForMultipleObjects(2, waitEvents, false, INFINITE);
	}
}


void View::ThreadProc()
{
	{
		PERFORMANCER;

		auto className = L"skCefContainerWindow";

		auto instance = (HINSTANCE)GetModuleHandle(nullptr);

		WNDCLASSEX wc;
		wc.cbSize = sizeof(WNDCLASSEX);
		wc.style = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc = (WNDPROC)View::WndProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = instance;
		wc.hIcon = nullptr;
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
		wc.lpszMenuName = nullptr;
		wc.lpszClassName = className;
		wc.hIconSm = nullptr;

		RegisterClassEx(&wc);

		auto style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_VISIBLE;
		auto exStyle = WS_EX_CLIENTEDGE;

		HWND parentHandle = nullptr;
		if (createParams.ParentId != -1)
		{
			auto parent = BrowserClient::GetInstance().GetViewManager().GetView(createParams.ParentId);
			if (parent != nullptr)
				parentHandle = parent->GetHandle();
		}

		bool isChildWindow = parentHandle != nullptr;
		if (isChildWindow)
		{
			style = WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
		}
		else
		{
			style |= WS_OVERLAPPED | WS_CLIPCHILDREN;
			exStyle |= WS_EX_APPWINDOW;
		}

		if (!Display.GetHasBorder())
		{
			style &= ~(WS_CAPTION | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_SYSMENU);
			exStyle &= ~(WS_EX_DLGMODALFRAME | WS_EX_CLIENTEDGE | WS_EX_STATICEDGE);
		}

		if (!createParams.IsVisible)
			style &= ~WS_VISIBLE;

		if (createParams.ShowInTaskbar)
			exStyle &= ~WS_EX_NOACTIVATE;
		else
			exStyle |= WS_EX_NOACTIVATE;

		auto rootView = BrowserClient::GetInstance().GetViewManager().GetRootView();
		if (rootView != nullptr && rootView.get() != this)
			parentHandle = rootView->GetHandle();

		CreateWindowEx(
			exStyle,
			className,
			L"",
			style,
			createParams.Left, createParams.Top, createParams.Width, createParams.Height,
			parentHandle, nullptr, instance, this);

		if (windowHandle == NULL)
			throw std::exception("Window creation failed.");

		if (createParams.IsVisible)
			ShowWindow(windowHandle, SW_SHOWNOACTIVATE);

		Display.UpdateStyles();

		if (!IsPopup())
		{
			CefBrowserSettings browserSettings;
			CefString(&browserSettings.default_encoding).FromString("UTF-8");
			browserSettings.file_access_from_file_urls = STATE_ENABLED;
			browserSettings.application_cache = STATE_ENABLED;
			browserSettings.local_storage = STATE_ENABLED;
			browserSettings.databases = STATE_ENABLED;
			browserSettings.javascript_close_windows = STATE_DISABLED;

			RECT windowRect;
			GetClientRect(windowHandle, &windowRect);

			CefWindowInfo windowInfo;
			windowInfo.SetAsChild(windowHandle, windowRect);

#ifndef USE_GLOBAL_REQUESTCONTEXT
			CefRequestContextSettings settings;
			auto requestContext = CefRequestContext::CreateContext(settings, new RequestContextHandler());

			CefRefPtr<CefSchemeHandlerFactory> factory = new SiteKioskSchemeHandlerFactory();
			requestContext->RegisterSchemeHandlerFactory("sk", "siteKiosk", factory);

			CefBrowserHost::CreateBrowser(windowInfo, BrowserClient::GetRefPtrInstance(), createParams.StartUrl, browserSettings, requestContext);
#else

			CefBrowserHost::CreateBrowser(windowInfo, BrowserClient::GetRefPtrInstance(), createParams.StartUrl, browserSettings, nullptr);
#endif

		}

		UpdateWindow(windowHandle);
	}

	MSG msg;
	while (GetMessage(&msg, windowHandle, 0, 0) > 0)
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}


void View::WaitForReady() const
{
	viewReady.Wait();
}
