#pragma once

#include "../Thread/ManualResetEvent.h"

#include "include/cef_app.h"
#include <mutex>
#include <vector>
#include <thread>
#include <functional>
#include <climits>


class BrowserRPCProcessor;
class DisplayComponent;
class LayoutComponent;
class InteractionComponent;
class RequestComponent;
class EventComponent;
class PingComponent;
class ErrorComponent;
class Cancelable;

struct CreateViewParams
{
	int Left;
	int Top;
	int Width;
	int Height;

	int MinWidth;
	int MinHeight;

	int MaxWidth;
	int MaxHeight;

	int ZIndex;

	bool ShowBorders;
	bool AllowInteraction;
	std::wstring StartUrl;
	bool IsClosable;
	bool IsVisible;
	bool ShowInTaskbar;
	int ContentFilterLevel;
	int ParentId;

	CreateViewParams()
		: Left(0)
		, Top(0)
		, Width(800)
		, Height(600)
		, MinWidth(0)
		, MinHeight(0)
		, MaxWidth(INT_MAX)
		, MaxHeight(INT_MAX)
		, ZIndex(0)
		, ShowBorders(true)
		, AllowInteraction(true)
		, IsClosable(true)
		, IsVisible(false)
		, ParentId(-1)
		, ShowInTaskbar(true)
		, ContentFilterLevel(0)
	{
	}
};

class View
{
	friend class ViewMessageLoop;
	friend class ViewManager;
public:
	View(int id);
	~View();

	void Init(const CreateViewParams& params);
	void Init(int creatorId, CefWindowInfo& windowInfo, CefRefPtr<CefBrowser> browser, CefBrowserSettings& settings);
	
	void SetBrowser(CefRefPtr<CefBrowser> browser);
	CefRefPtr<CefBrowser> GetBrowser();

	int GetId();
	std::shared_ptr<BrowserRPCProcessor> GetBrowserRPCProcessor();
	ManualResetEvent& GetViewReadyEvent();

	std::wstring GetString(const std::wstring& stringId);

	void Close();
	void Close(bool isUserInstigated);

	bool IsMainView() const { return isMainView; }

	HWND GetHandle() { return windowHandle; }

	void SetAllowClose() { allowClose = true; }
	bool IsPopup() const  { return popupCreatorId != -1; }
	int GetPopupCreatorId() const { return popupCreatorId; }

	void WaitForReady() const;

private:
	void CreateContainerWindow(bool isPopupWindow);
	void ThreadProc();

	CreateViewParams createParams;
	HWND windowHandle;
	std::unique_ptr<std::thread> windowThread;
	bool allowClose;

private:
	ManualResetEvent viewReady;
	ManualResetEvent browserReady;
	CefRefPtr<CefBrowser> browser;
	int id;
	int popupCreatorId;
	bool isClosing;

	std::recursive_mutex mutex;

private:
	std::unique_ptr<LayoutComponent> layoutComponent;
	std::unique_ptr<DisplayComponent> displayComponent;
	std::unique_ptr<InteractionComponent> interactionComponent;
	std::unique_ptr<RequestComponent> requestComponent;
	std::unique_ptr<EventComponent> eventComponent;
	std::unique_ptr<PingComponent> pingComponent;
	std::unique_ptr<ErrorComponent> errorComponent;

public:
	LayoutComponent& Layout;
	DisplayComponent& Display;
	InteractionComponent& Interaction;
	RequestComponent& Requests;
	EventComponent& Events;
	PingComponent& Pinger;
	ErrorComponent& Error;

	bool isMainView;

public:
	static LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
	LRESULT OnMessage(UINT msg, WPARAM wparam, LPARAM lparam);
	LRESULT OnCreate(WPARAM wparam, LPARAM lparam);
	LRESULT OnPointerActivate(WPARAM wparam, LPARAM lparam);
	LRESULT OnEraseBackground(WPARAM wparam, LPARAM lparam);
	LRESULT OnActivate(WPARAM wparam, LPARAM lparam);
	LRESULT OnKillFocus(WPARAM wparam, LPARAM lparam);
	LRESULT OnSetFocus(WPARAM wparam, LPARAM lparam);
	LRESULT OnMove(WPARAM wparam, LPARAM lparam);
	LRESULT OnSize(WPARAM wparam, LPARAM lparam);
	LRESULT OnClose(WPARAM wparam, LPARAM lparam);
	LRESULT OnDestroy(WPARAM wparam, LPARAM lparam);
	LRESULT OnGetMinMaxInfo(WPARAM wparam, LPARAM lparam);
};

