#include "ViewManager.h"
#include "View.h"
#include "Components\LayoutComponent.h"
#include "Components\InteractionComponent.h"
#include "Components\EventComponent.h"
#include "Components\PingComponent.h"
#include "Components\DisplayComponent.h"
#include "../Browser/BrowserClient.h"
#include "../Messages/WebAppRPCProcessor.h"
#include "../Utils/Profiler.h"
#include "../Utils/CommandLine.h"
#include "../Utils/StringHelper.h"
#include "../Utils/ThreadHelper.h"

#include <algorithm>
#include <json/json.h>


ViewManager::ViewManager()
	: nextViewId(0)
	, isPinging(false)
	, lastFocusedView(0)
	, viewsExisting(false)
{
}


ViewManager::~ViewManager()
{
}


std::shared_ptr<View> ViewManager::Create()
{
	std::shared_ptr<View> view;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);
		view = std::shared_ptr<View>(new View(nextViewId++));
		views.push_back(view);
	}

	return view;
}


std::shared_ptr<View> ViewManager::Create(const CreateViewParams& params)
{
	auto view = Create();
	view->Init(params);

	return view;
}


void ViewManager::CreateFromPopup(std::shared_ptr<View> creator, CefWindowInfo& windowInfo, CefRefPtr<CefBrowser> browser, CefBrowserSettings& settings)
{
	std::shared_ptr<View> view;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);
		view = std::shared_ptr<View>(new View(nextViewId++));
		views.push_back(view);
	}

	view->Init(creator->GetId(), windowInfo, browser, settings);
}


void ViewManager::CloseView(int viewId, const std::function<void()>& closed)
{
	auto view = GetView(viewId);
	
	if (view != nullptr)
	{
		{
			std::unique_lock<std::recursive_mutex> lock(mutex);

			auto iter = registeredViewClosedEvents.find(viewId);
			if (iter == registeredViewClosedEvents.end())
			{
				std::vector<std::function<void()>> events;
				registeredViewClosedEvents[viewId] = events;
			}
			registeredViewClosedEvents[viewId].push_back(closed);
		}
		view->Close();
	}
	else
		closed();
}


void ViewManager::RemoveView(std::shared_ptr<View> view)
{
	bool hasViews = false;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);

		auto iter = std::find_if(views.begin(), views.end(), [&](std::shared_ptr<View> & otherView) -> bool
		{
			auto equal = otherView == view;
			return equal;
		});

		if (iter != views.end())
		{
			views.erase(iter);
		}

		hasViews = !views.empty();
	}

	view->Close(false);

	if (!hasViews)
	{
		viewsExisting.Reset();
		if (shutDownCallback)
			shutDownCallback();
	}

	if (!HasViews())
	{
		OutputDebugString(L"All views have been closed, shutting down the CefApp.");
		// All browser windows have closed. Quit the application message loop.
		BrowserClient::GetInstance().Quit();
	}
}

void ViewManager::OnViewDestructed(int viewId)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	auto iter = registeredViewClosedEvents.find(viewId);
	if (iter != registeredViewClosedEvents.end())
	{
		for (const auto& entry : iter->second)
			entry();
		registeredViewClosedEvents.erase(iter);
	}
}


std::shared_ptr<View> ViewManager::GetView(CefRefPtr<CefBrowser> browser)
{
	if (browser == nullptr)
		return nullptr;

	std::unique_lock<std::recursive_mutex> lock(mutex);

	auto viewIterator = std::find_if(views.begin(), views.end(), [&](std::shared_ptr<View> & view) -> bool
	{
		auto otherBrowser = view->browser;
		if (otherBrowser == nullptr)
			return false;
		return browser->IsSame(view->browser);
	});
	if (viewIterator == views.end())
		return nullptr;
	return *viewIterator;
}


std::shared_ptr<View> ViewManager::GetView(int id)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	auto viewIterator = std::find_if(views.begin(), views.end(), [&](std::shared_ptr<View> & view) -> bool
	{
		return view->GetId() == id;
	});
	if (viewIterator == views.end())
		return nullptr;
	return *viewIterator;
}



std::shared_ptr<View> ViewManager::GetView(HWND hwnd)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	auto viewIterator = std::find_if(views.begin(), views.end(), [&](std::shared_ptr<View> & view) -> bool
	{
		return view->GetHandle() == hwnd;
	});
	if (viewIterator == views.end())
		return nullptr;
	return *viewIterator;
}


std::shared_ptr<View> ViewManager::GetRootView(int id)
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	std::shared_ptr<View> root;

	while (true)
	{
		auto viewIterator = std::find_if(views.begin(), views.end(), [&](std::shared_ptr<View> & view) -> bool
		{
			return view->GetId() == id;
		});
		if (viewIterator == views.end())
			break;
		root = *viewIterator;
		id = root->Layout.GetParent();
	}
	return root;
}


std::shared_ptr<View> ViewManager::GetRootView()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	if (views.empty())
		return nullptr;
	return views.front();
}


ManualResetEvent& ViewManager::GetViewExistingEvent()
{
	return viewsExisting;
}


void ViewManager::Ping()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);

	if (isPinging)
		return;

	if (views.empty()) 
	{
		BrowserClient::GetInstance().GetRPCProcessor().CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"pong").Run();
		return;
	}

	isPinging = true;

	receivedPongResponses = 0;
	int expectedPongs = views.size();
	auto receivedPongCallback = [&, expectedPongs]()
	{
		receivedPongResponses++;
		if (receivedPongResponses >= expectedPongs)
		{
			BrowserClient::GetInstance().GetRPCProcessor().CreateRequest(WebAppRPCProcessor::AppCommandPrefix, L"pong").Run();
			isPinging = false;
		}
	};
	
	for (auto &view : views)
		view->Pinger.Ping(receivedPongCallback);
}


void ViewManager::OnRendererCreated(int browserId)
{
	std::vector<std::shared_ptr<View>> tmpViews;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);
		tmpViews = views;
	}
	

	for (auto &view : tmpViews)
	{
		auto browser = view->browser;
		if (browser != nullptr && browser->GetIdentifier() == browserId)
			view->Events.OnRendererCreated();
	}
}


bool ViewManager::HasViews()
{
	std::unique_lock<std::recursive_mutex> lock(mutex);
	return !views.empty();
}


void ViewManager::Shutdown(std::function<void()> callback)
{
	shutDownCallback = callback;
	std::vector<std::shared_ptr<View>> tmpViews;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);
		tmpViews = views;
	}

	if (tmpViews.empty())
		shutDownCallback();
	else
		for (int i = tmpViews.size() - 1; i >= 0; i--)
			tmpViews[i]->Close();
}


void ViewManager::SetLastFocusedView(int viewId)
{
	lastFocusedView = viewId;
}


void ViewManager::OnZOrderChanged(int viewId)
{
	std::vector<std::pair<std::shared_ptr<View>, int>> siblings;
	{
		std::unique_lock<std::recursive_mutex> lock(mutex);

		auto view = GetView(viewId);
		if (view == nullptr)
			return;

		siblings.push_back(std::make_pair(view, 0));

		auto parent = view->Layout.GetParent();

		for (auto& otherView : views)
			if (otherView != view && otherView->Layout.GetParent() == parent)
				siblings.push_back(std::make_pair(otherView, 0));


		int zIndex = 0;
		auto window = GetTopWindow(GetParent(view->GetHandle()));
		while (true)
		{
			zIndex++;

			for (auto& sibling : siblings)
				if (sibling.first->GetHandle() == window)
					sibling.second = zIndex;

			window = GetNextWindow(window, GW_HWNDNEXT);
			if (window == nullptr)
				break;
		}
	}

	std::sort(siblings.begin(), siblings.end(), [](const std::pair<std::shared_ptr<View>, int>& a, const std::pair<std::shared_ptr<View>, int>& b)
	{
		auto aIndex = a.second;
		auto bIndex = b.second;
		return aIndex > bIndex;
	});

	std::sort(siblings.begin(), siblings.end(), [](const std::pair<std::shared_ptr<View>, int>& a, const std::pair<std::shared_ptr<View>, int>& b)
	{
		auto aIndex = a.first->Layout.GetZIndex();
		auto bIndex = b.first->Layout.GetZIndex();
		return aIndex < bIndex;
	});

	std::shared_ptr<View> previousSibling = nullptr;
	for (auto& sibling : siblings)
	{
		auto view = sibling.first;
		auto handle = view->GetHandle();
		if (view->Layout.GetZIndex() >= 1000)
			::SetWindowPos(handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE | SWP_NOOWNERZORDER);
		else
			::SetWindowPos(handle, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE | SWP_NOOWNERZORDER);
	}
}
