#include "View.h"

#include "Components/InteractionComponent.h"
#include "Components/EventComponent.h"
#include "Components/LayoutComponent.h"

#include "../Browser/BrowserClient.h"
#include "../Utils/StringHelper.h"


LRESULT CALLBACK View::WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	View* view = nullptr;
	if (msg == WM_CREATE)
	{
		auto createStruct = (LPCREATESTRUCT)lParam;
		SetWindowLong(hwnd, GWLP_USERDATA, (long)createStruct->lpCreateParams);
		view = (View*)createStruct->lpCreateParams;
		view->windowHandle = hwnd;
	}
	else
		view = (View *)GetWindowLong(hwnd, GWLP_USERDATA);

	if (view == nullptr)
		return DefWindowProc(hwnd, msg, wParam, lParam);

	return view->OnMessage(msg, wParam, lParam);
}


LRESULT View::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_CREATE:
		return OnCreate(wParam, lParam);

	case WM_POINTERACTIVATE:
		return OnPointerActivate(wParam, lParam);

	case WM_ERASEBKGND:
		return OnEraseBackground(wParam, lParam);

	case WM_ACTIVATE:
		return OnActivate(wParam, lParam);

	case WM_KILLFOCUS:
		return OnKillFocus(wParam, lParam);

	case WM_SETFOCUS:
		return OnSetFocus(wParam, lParam);

	case WM_MOVE:
		return OnMove(wParam, lParam);

	case WM_SIZE:
		return OnSize(wParam, lParam);

	case WM_CLOSE:
		return OnClose(wParam, lParam);

	case WM_DESTROY:
		return OnDestroy(wParam, lParam);

	case WM_GETMINMAXINFO:
		return OnGetMinMaxInfo(wParam, lParam);

	default:
		return DefWindowProc(windowHandle, msg, wParam, lParam);
	}

	return 0;
}


LRESULT View::OnCreate(WPARAM wparam, LPARAM lparam)
{
	viewReady.Set();

	return 0;
}


LRESULT View::OnPointerActivate(WPARAM wParam, LPARAM lParam)
{
	if (browser)
	{
		::SetFocus(browser->GetHost()->GetWindowHandle());
		browser->GetHost()->SetFocus(true);
	}
	return PA_NOACTIVATE;
}


LRESULT View::OnEraseBackground(WPARAM wparam, LPARAM lparam)
{
	if (browser)
		return 0;
	return 1;
}


LRESULT View::OnActivate(WPARAM wParam, LPARAM lParam)
{
	if (wParam == WA_INACTIVE)
		Events.OnDeactivated();
	else
		Events.OnActivated();

	return 0;
}


LRESULT View::OnKillFocus(WPARAM wParam, LPARAM lParam)
{
	if (browser)
		browser->GetHost()->SetFocus(false);
	return 0;
}


LRESULT View::OnSetFocus(WPARAM wParam, LPARAM lParam)
{
	if (browser)
		browser->GetHost()->SetFocus(true);
	return 0;
}


LRESULT View::OnMove(WPARAM wParam, LPARAM lParam)
{
	if (browser)
	{
		// needed to update JS: window.screenX and stuff
		browser->GetHost()->NotifyMoveOrResizeStarted();
		browser->GetHost()->WasResized();
	}
	return 0;
}


LRESULT View::OnSize(WPARAM wParam, LPARAM lParam)
{
	if (browser)
	{
		SetWindowPos(browser->GetHost()->GetWindowHandle(), nullptr, 0, 0, LOWORD(lParam), HIWORD(lParam), SWP_ASYNCWINDOWPOS | SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOOWNERZORDER);
		// needed to update JS: window.screenX and stuff
		browser->GetHost()->NotifyMoveOrResizeStarted();
		browser->GetHost()->WasResized();
	}
	return 0;
}


LRESULT View::OnClose(WPARAM wParam, LPARAM lParam)
{
	viewReady.Wait();

	if (!allowClose)
	{
		if (!Interaction.GetIsCloseable() && !BrowserClient::GetInstance().IsClosing())
			return 1;

		if (browser)
			browser->GetHost()->CloseBrowser(false);
	}
	else
	{
		DestroyWindow(windowHandle);
	}
	return 0;
}


LRESULT View::OnDestroy(WPARAM wParam, LPARAM lParam)
{
	SetWindowLong(windowHandle, GWLP_USERDATA, 0);
	PostQuitMessage(0);
	return 0;
}


LRESULT View::OnGetMinMaxInfo(WPARAM wParam, LPARAM lParam)
{
	MINMAXINFO* minMaxInfo = (MINMAXINFO*)lParam;
	auto maxSize = Layout.GetMaxSize();
	auto minSize = Layout.GetMinSize();
	minMaxInfo->ptMinTrackSize.x = minSize.cx;
	minMaxInfo->ptMinTrackSize.y = minSize.cy;
	minMaxInfo->ptMaxTrackSize.x = maxSize.cx;
	minMaxInfo->ptMaxTrackSize.y = maxSize.cy;

	return 0;
}
